import torch
from torch.nn import Linear, ReLU, LeakyReLU
from torch import nn
from torch.utils.data import Dataset, DataLoader
import os
from .utils import cal_num_bins
from torch.optim import lr_scheduler
import sys

class Semi_encoding_multiple(torch.nn.Module):
    """
    Model for combined features
    """
    def __init__(self, num):
        super(Semi_encoding_multiple, self).__init__()
        self.encoder1 = torch.nn.Sequential(
            Linear(num, 512),
            nn.BatchNorm1d(512),
            LeakyReLU(),
            nn.Dropout(0.2),
            Linear(512, 512),
            nn.BatchNorm1d(512),
            LeakyReLU(),
            nn.Dropout(0.2),
            Linear(512, 100),
        )

        self.decoder1 = torch.nn.Sequential(
            Linear(100, 512),
            nn.BatchNorm1d(512),
            LeakyReLU(),
            nn.Dropout(0.2),
            Linear(512, 512),
            nn.BatchNorm1d(512),
            LeakyReLU(),
            nn.Dropout(0.2),
            Linear(512, num),
            nn.Sigmoid(),
        )

    def forward(self, input1, input2):
        return self.encoder1(input1), self.encoder1(input2)

    def decoder(self, input1, input2):
        return self.decoder1(input1), self.decoder1(input2)

    def embedding(self, input):
        return self.encoder1(input)


class Semi_encoding_single(torch.nn.Module):
    """
    Model for k-mer features
    """
    def __init__(self, num):
        super(Semi_encoding_single, self).__init__()
        self.encoder1 = torch.nn.Sequential(
            Linear(num, 512),
            nn.BatchNorm1d(512),
            LeakyReLU(),
            nn.Dropout(0.2),
            Linear(512, 512),
            nn.BatchNorm1d(512),
            LeakyReLU(),
            nn.Dropout(0.2),
            Linear(512, 100),
        )

        self.decoder1 = torch.nn.Sequential(
            Linear(100, 512),
            nn.BatchNorm1d(512),
            LeakyReLU(),
            nn.Dropout(0.2),
            Linear(512, 512),
            nn.BatchNorm1d(512),
            LeakyReLU(),
            nn.Dropout(0.2),
            Linear(512, num),
            nn.Softmax(dim=1),
        )

    def forward(self, input1, input2):
        return self.encoder1(input1), self.encoder1(input2)

    def decoder(self, input1, input2):
        return self.decoder1(input1), self.decoder1(input2)

    def embedding(self, input):
        return self.encoder1(input)


def loss_function(embedding1, embedding2, label, raw_x_1,
                  raw_x_2, decoder_x_1, decoder_x_2, is_label=True):
    relu = torch.nn.ReLU()
    mse_loss = torch.nn.MSELoss()
    d = torch.norm(embedding1 - embedding2, p=2, dim=1)
    square_pred = torch.square(d)
    margin_square = torch.square(relu(1 - d))

    if is_label:
        supervised_loss = torch.mean(
            label * square_pred + (1 - label) * margin_square)
        unsupervised_loss = 0.5 * \
            mse_loss(decoder_x_1, raw_x_1) + 0.5 * \
            mse_loss(decoder_x_2, raw_x_2)
        loss = supervised_loss + unsupervised_loss
        return loss, supervised_loss, unsupervised_loss

    else:
        unsupervised_loss = 0.5 * \
            mse_loss(decoder_x_1, raw_x_1) + 0.5 * \
            mse_loss(decoder_x_2, raw_x_2)
        return unsupervised_loss


class feature_Dataset(Dataset):
    def __init__(self, embedding1_lists, embedding2_lists, label_lists):
        self.embedding1_lists = embedding1_lists
        self.embedding2_lists = embedding2_lists
        self.label_lists = label_lists

    def __getitem__(self, item):
        return self.embedding1_lists[item], self.embedding2_lists[item], self.label_lists[item]

    def __len__(self):
        return len(self.embedding1_lists)


class unsupervised_feature_Dataset(Dataset):
    def __init__(self, embedding1_lists, embedding2_lists):
        self.embedding1_lists = embedding1_lists
        self.embedding2_lists = embedding2_lists

    def __getitem__(self, item):
        return self.embedding1_lists[item], self.embedding2_lists[item]

    def __len__(self):
        return len(self.embedding1_lists)


def train(out, contig_fastas, binned_lengths, logger, datas, data_splits, cannot_links, is_combined=True,
          batchsize=2048, epoches=20, device=None, num_process = 8, mode = 'single', orf_finder = 'prodigal'):
    """
    Train model from one sample(--mode single) or several samples(--mode several)
    """
    from tqdm import tqdm
    import pandas as pd

    train_data = pd.read_csv(datas[0], index_col=0).values
    if not is_combined:
        train_data_input = train_data[:, 0:136]
    else:
        train_data_input = train_data


    torch.set_num_threads(num_process)

    logger.info('Training model...')

    if not is_combined:
        model = Semi_encoding_single(train_data_input.shape[1]).to(device)
    else:
        model = Semi_encoding_multiple(train_data_input.shape[1]).to(device)

    optimizer = torch.optim.Adam(model.parameters(), lr=1e-3)
    scheduler = lr_scheduler.StepLR(optimizer, step_size=1, gamma=0.9)

    for epoch in tqdm(range(epoches)):
        for data_index in range(len(contig_fastas)):
            seed = cal_num_bins(
                contig_fastas[data_index],
                binned_length=binned_lengths[data_index],
                num_process=num_process,
                output=os.path.join(out, 'sample{}'.format(data_index)),
                orf_finder=orf_finder)
            if epoch == 0:
                logger.debug('Generate training data of {}:'.format(data_index))

            # #generate two inputs
            train_input_1 = []
            train_input_2 = []
            train_labels = []

            # cannot link

            if not os.path.getsize(cannot_links[data_index]):
                sys.stderr.write(
                    f"Error: Cannot-link file is empty!\n")
                sys.exit(1)

            cannot_link = pd.read_csv(cannot_links[data_index], sep=',',
                                      header=None).values
            data = pd.read_csv(datas[data_index], index_col=0)
            data.index = data.index.astype(str)
            data_split = pd.read_csv(data_splits[data_index], index_col=0)

            if mode == 'several':
                if data.shape[1] != 138 or data_split.shape[1] != 136:
                    sys.stderr.write(
                        f"Error: training mode with several only used in single-sample binning!\n")
                    sys.exit(1)

            namelist = data.index.tolist()
            mapObj = dict(zip(namelist, range(len(namelist))))
            train_data = data.values
            train_data_must_link = data_split.values

            if not is_combined:
                train_data_input = train_data[:, 0:136]
                train_data_split_input = train_data_must_link
            else:
                train_data_input = train_data
                train_data_split_input = train_data_must_link

            # can not link from contig annotation
            for link in cannot_link:
                train_input_1.append(train_data_input[mapObj[str(link[0])]])
                train_input_2.append(train_data_input[mapObj[str(link[1])]])
                train_labels.append(0)

            # cannot link from bin seed
            if seed is not None:
                for i in range(len(seed)):
                    for j in range(i + 1, len(seed)):
                        train_input_1.append(
                            train_data_input[mapObj[str(seed[i])]])
                        train_input_2.append(
                            train_data_input[mapObj[str(seed[j])]])
                        train_labels.append(0)

            # must link from breaking up
            for i in range(0, len(train_data_must_link), 2):
                train_input_1.append(train_data_split_input[i])
                train_input_2.append(train_data_split_input[i + 1])
                train_labels.append(1)

            if epoch == 0:
                logger.debug(
                    'Number of must-link pairs: {}'.format(train_labels.count(1)))
                logger.debug(
                    'Number of cannot-link pairs: {}'.format(train_labels.count(0)))

            dataset = feature_Dataset(train_input_1, train_input_2,
                                      train_labels)
            train_loader = DataLoader(
                dataset=dataset,
                batch_size=batchsize,
                shuffle=True,
                num_workers=0)

            unlabeled_x = train_data_input
            unlabeled_train_input1 = []
            unlabeled_train_input2 = []
            for i in range(len(unlabeled_x)):
                unlabeled_train_input1.append(unlabeled_x[i])
                unlabeled_train_input2.append(unlabeled_x[i])

            dataset_unlabeled = unsupervised_feature_Dataset(
                unlabeled_train_input1, unlabeled_train_input2)
            train_loader_unlabeled = DataLoader(
                dataset=dataset_unlabeled,
                batch_size=batchsize,
                shuffle=True,
                num_workers=0)

            for train_input1, train_input2, train_label in train_loader:
                model.train()
                train_input1 = train_input1.to(device)
                train_input2 = train_input2.to(device)
                train_label = train_label.to(device)
                embedding1, embedding2 = model.forward(
                    train_input1.float(), train_input2.float())
                decoder1, decoder2 = model.decoder(embedding1, embedding2)
                optimizer.zero_grad()
                loss, supervised_loss, unsupervised_loss = loss_function(embedding1.double(), embedding2.double(),
                                                                         train_label.double(), train_input1.double(),
                                                                         train_input2.double(), decoder1.double(),
                                                                         decoder2.double())
                loss = loss.to(device)
                loss.backward()
                optimizer.step()

            for unlabeled_train_input1, unlabeled_train_input2 in train_loader_unlabeled:
                model.train()
                unlabeled_train_input1 = unlabeled_train_input1.to(device)
                unlabeled_train_input2 = unlabeled_train_input2.to(device)
                embedding1, embedding2 = model.forward(
                    unlabeled_train_input1.float(), unlabeled_train_input1.float())
                decoder1, decoder2 = model.decoder(embedding1, embedding2)
                optimizer.zero_grad()
                loss = loss_function(embedding1.double(), embedding2.double(),
                                     None, unlabeled_train_input1.double(),
                                     unlabeled_train_input2.double(), decoder1.double(),
                                     decoder2.double(), is_label=False).to(device)
                loss.backward()
                optimizer.step()
        scheduler.step()

    logger.info('Training finished.')
    torch.save(model, os.path.join(out, 'model.h5'))

    return model



import numpy as np
from SemiBin.cluster import cal_kl

def slow_cal_kl(m, v):
    """
    Calculate KL divergence
    """
    m = np.clip(m, 1e-6, None)
    v = np.clip(v, 1.0, None)
    m1 = m.reshape(1, len(m))
    m2 = m.reshape(len(m), 1)

    v1 = v.reshape(1, len(v))
    v2 = v.reshape(len(v), 1)

    value = np.log(np.sqrt(v1 / v2)) +  np.divide(np.square(m1 - m2) + v2, 2 * v1) - 0.5

    return np.clip(value, 1e-6, 1 - 1e-6)

def test_cal_kl():
    for _ in range(32):
        m = np.random.rand(128)
        v = np.random.rand(128)
        assert np.allclose(slow_cal_kl(m, v), cal_kl(m ,v, use_ne='yes'))
        assert np.allclose(slow_cal_kl(m, v), cal_kl(m ,v, use_ne='no'))
        assert np.allclose(slow_cal_kl(m, v), cal_kl(m ,v, use_ne='auto'))



import os
import subprocess
import contextlib
import multiprocessing
from .atomicwrite import atomic_write
import tempfile
import sys
import random
import shutil

from .fasta import fasta_iter

def validate_normalize_args(logger, args):
    '''Validate and normalize command line arguments'''
    def expect_file(f):
        if f is not None:
            if not os.path.exists(f):
                sys.stderr.write(
                    f"Error: Expected file '{f}' does not exist\n")
                sys.exit(1)

    def expect_file_list(fs):
        if fs is not None:
            for f in fs:
                expect_file(f)

    if args.cmd != 'download_GTDB' and args.cmd != 'check_install' and args.cmd != 'concatenate_fasta':
        if args.num_process == 0:
            args.num_process = multiprocessing.cpu_count()
            logger.info(f'Setting number of CPUs to {args.num_process}')

        if args.num_process > multiprocessing.cpu_count():
            args.num_process = multiprocessing.cpu_count()

    if args.cmd in ['single_easy_bin', 'multi_easy_bin', 'train', 'bin', 'train_self']:
        if args.cmd != 'train_self':
            if args.orf_finder not in ['prodigal', 'fraggenescan']:
                sys.stderr.write(
                    f"Error: SemiBin only supports prodigal or fraggenescan as the ORF finder (--orf_finder option).\n")
                sys.exit(1)
        if args.engine not in ['auto', 'gpu', 'cpu']:
            sys.stderr.write(
                f"Error: You need to specify the engine in auto/gpu/cpu.\n")
            sys.exit(1)

    if args.cmd == 'generate_cannot_links':
        if args.GTDB_reference is not None:
            expect_file(args.GTDB_reference)
        expect_file(args.contig_fasta)
        expect_file_list(args.taxonomy_results_fname)

    if args.cmd == 'generate_sequence_features_single':
        expect_file(args.contig_fasta)
        expect_file_list(args.bams)

    if args.cmd == 'generate_sequence_features_multi':
        expect_file(args.contig_fasta)
        expect_file_list(args.bams)

    if args.cmd in ['train', 'train_self']:
        if args.mode == 'single':
            if len(args.data) > 1:
                sys.stderr.write(
                    f"Error: Expected one data.csv file with single mode.\n")
                sys.exit(1)

            if len(args.data_split) > 1:
                sys.stderr.write(
                    f"Error: Expected one data_split.csv file with single mode.\n")
                sys.exit(1)
            if args.cmd == 'train':
                if len(args.contig_fasta) > 1:
                    sys.stderr.write(
                        f"Error: Expected one fasta file with single mode.\n")
                    sys.exit(1)

                if len(args.cannot_link) > 1:
                    sys.stderr.write(
                        f"Error: Expected one cannot.txt file with single mode.\n")
                    sys.exit(1)
                expect_file(args.cannot_link[0])
                expect_file(args.contig_fasta[0])

            expect_file(args.data[0])
            expect_file(args.data_split[0])

        elif args.mode == 'several':
            if args.cmd == 'train':
                assert len(args.contig_fasta) == len(args.data) == len(args.data_split) == len(args.cannot_link), 'Must input same number of fasta, data, data_split, cannot files!'
                expect_file_list(args.cannot_link)
                expect_file_list(args.contig_fasta)
            else:
                assert len(args.data) == len(args.data_split) , 'Must input same number of data, data_split!'

            expect_file_list(args.data)
            expect_file_list(args.data_split)


        else:
            sys.stderr.write(
                f"Error: Please use training mode with [single/several].\n")
            sys.exit(1)

    if args.cmd == 'bin':
        if args.environment is None and args.model_path is None:
            sys.stderr.write(
                f"Error: Please choose input a model path or use our built-in model for [human_gut/dog_gut/ocean/soil/cat_gut/human_oral/mouse_gut/pig_gut/built_environment/wastewater/chicken_caecum/global].\n")
            sys.exit(1)
        if args.environment is not None and args.model_path is not None:
            sys.stderr.write(
                f"Error: Please choose input a model path or use our built-in model for [human_gut/dog_gut/ocean/soil/cat_gut/human_oral/mouse_gut/pig_gut/built_environment/wastewater/chicken_caecum/global].\n")
            sys.exit(1)
        if args.environment is not None:
            # This triggers checking that the environment is valid
            get_model_path(args.environment)
        if args.model_path is not None:
            expect_file(args.model_path)
        expect_file(args.contig_fasta)
        expect_file(args.data)

    if args.cmd == 'single_easy_bin':
        if args.GTDB_reference is not None:
            expect_file(args.GTDB_reference)
        expect_file(args.contig_fasta)
        expect_file_list(args.bams)

        if args.environment is not None:
            # This triggers checking that the environment is valid
            get_model_path(args.environment)

        if args.training_type not in ['semi', 'self']:
            sys.stderr.write(
                f"Error: You need to specify the training algorithm in semi/self.\n")
            sys.exit(1)

    if args.cmd == 'multi_easy_bin':
        if args.GTDB_reference is not None:
            expect_file(args.GTDB_reference)
        expect_file(args.contig_fasta)
        expect_file_list(args.bams)

        if args.training_type not in ['semi', 'self']:
            sys.stderr.write(
                f"Error: You need to specify the training algorithm in semi/self.\n")
            sys.exit(1)

    if args.cmd == 'concatenate_fasta':
        expect_file_list(args.contig_fasta)
        contig_name = []
        for contig in args.contig_fasta:
            contig_name.append(os.path.basename(contig).split('.')[0])
        if len(set(contig_name)) != len(contig_name):
            sys.stderr.write(
                f"Error: Make sure that every contig file have different names.\n")
            sys.exit(1)



def get_must_link_threshold(contig_lens):
    """
    calculate the threshold length for must link breaking up
    """
    import numpy as np
    contig_lens = np.array(contig_lens)
    contig_lens.sort()
    frac = np.cumsum(contig_lens)/np.sum(contig_lens)
    ix = np.argmax(frac > 0.02) # argmax finds first True element
    threshold = contig_lens[ix]
    return np.clip(threshold, 4000, None)

def parse_mmseqs(mmseqs_result):
    species_result = mmseqs_result.query('rank_name == "species" and score > 0.95').values
    genus_result = mmseqs_result.query('rank_name == "genus" and score > 0.80').values

    cannot_link_species = []
    for i in range(len(species_result)):
        for j in range(i + 1, len(species_result)):
            if species_result[i][2] != species_result[j][2]:
                cannot_link_species.append(
                    (species_result[i][0], species_result[j][0]))

    cannot_link_genus = []
    for i in range(len(genus_result)):
        for j in range(i + 1, len(genus_result)):
            if genus_result[i][2] != genus_result[j][2]:
                cannot_link_genus.append(
                    (genus_result[i][0], genus_result[j][0]))

    cannot_link_mix = []
    for i in range(len(species_result)):
        genus_name = species_result[i][4].split(';')[-2]
        for j in range(len(genus_result)):
            if genus_name != genus_result[j][2]:
                cannot_link_mix.append(
                    (species_result[i][0], genus_result[j][0]))

    return cannot_link_species, cannot_link_genus, cannot_link_mix


def generate_cannot_link(mmseqs_path, namelist, num_threshold, output,sample):
    import itertools
    import numpy as np
    import pandas as pd
    mmseqs_result = pd.read_csv(mmseqs_path,
                        sep='\t',
                        header=None,
                        names=['contig_name', 'taxon_ID', 'rank_name',
                                'scientific_name', 'temp_1', 'temp_2', 'temp_3', 'score', 'lineage'],
                        usecols=['contig_name', 'rank_name', 'scientific_name', 'score', 'lineage'])

    mmseqs_result['contig_name'] = mmseqs_result['contig_name'].astype(str)
    mmseqs_result = mmseqs_result[mmseqs_result['contig_name'].isin(
        namelist)]
    cannot_link_species, cannot_link_genus, cannot_link_mix = parse_mmseqs(
        mmseqs_result)
    num_whole_data = np.clip(1000 * num_threshold, None, 4_000_000)
    max_num_mix = int(num_whole_data / 8)
    max_num_genus = int(max_num_mix / 2)
    max_num_species = num_whole_data - max_num_mix - max_num_genus
    if len(cannot_link_mix) > max_num_mix:
        cannot_link_mix = random.sample(cannot_link_mix, max_num_mix)

    if len(cannot_link_genus) > max_num_genus:
        cannot_link_genus = random.sample(cannot_link_genus, max_num_genus)

    if len(cannot_link_species) > max_num_species:
        cannot_link_species = random.sample(cannot_link_species, max_num_species)

    with open(os.path.join(output, f'{sample}.txt'), 'w') as out_cannot_link:
        for cannot in itertools.chain(
                cannot_link_species,
                cannot_link_genus,
                cannot_link_mix):
            out_cannot_link.write(f'{cannot[0]},{cannot[1]}\n')


normalize_marker_trans__dict = {
    'TIGR00388': 'TIGR00389',
    'TIGR00471': 'TIGR00472',
    'TIGR00408': 'TIGR00409',
    'TIGR02386': 'TIGR02387',
}

def get_marker(hmmout, fasta_path=None, min_contig_len=None, multi_mode=False, orf_finder = None):
    import pandas as pd
    data = pd.read_table(hmmout, sep=r'\s+',  comment='#', header=None,
                         usecols=(0,3,5,15,16), names=['orf', 'gene', 'qlen', 'qstart', 'qend'])
    if not len(data):
        return []
    data['gene'] = data['gene'].map(lambda m: normalize_marker_trans__dict.get(m , m))
    qlen = data[['gene','qlen']].drop_duplicates().set_index('gene')['qlen']

    def contig_name(ell):
        if orf_finder == 'prodigal':
            contig,_ = ell.rsplit( '_', 1)
        else:
            contig,_,_,_ = ell.rsplit( '_', 3)
        return contig

    data = data.query('(qend - qstart) / qlen > 0.4').copy()
    data['contig'] = data['orf'].map(contig_name)
    if min_contig_len is not None:
        contig_len = {h:len(seq) for h,seq in fasta_iter(fasta_path)}
        data = data[data['contig'].map(lambda c: contig_len[c] >= min_contig_len)]
    data = data.drop_duplicates(['gene', 'contig'])

    def extract_seeds(vs, sel):
        vs = vs.sort_values()
        median = vs[len(vs) //2]

        # the original version broke ties by picking the shortest query, so we
        # replicate that here:
        candidates = vs.index[vs == median]
        c = qlen.loc[candidates].idxmin()
        r = list(sel.query('gene == @c')['contig'])
        r.sort()
        return r


    if multi_mode:
        data['bin'] = data['orf'].str.split('.', n=0, expand=True)[0]
        counts = data.groupby(['bin', 'gene'])['orf'].count()
        res = {}
        for b,vs in counts.groupby(level=0):
            cs = extract_seeds(vs.droplevel(0), data.query('bin == @b', local_dict={'b':b}))
            res[b] = [c.split('.',1)[1] for c in cs]
        return res
    else:
        counts = data.groupby('gene')['orf'].count()
        return extract_seeds(counts, data)

def prodigal(contig_file, contig_output):
        with open(contig_output + '.out', 'w') as prodigal_out_log:
            subprocess.check_call(
                ['prodigal',
                 '-i', contig_file,
                 '-p', 'meta',
                 '-q',
                 '-m', # See https://github.com/BigDataBiology/SemiBin/issues/87
                 '-a', contig_output
                 ],
                stdout=prodigal_out_log,
            )

def run_prodigal(fasta_path, num_process, output):
    from .error import LoggingPool

    contigs = {}
    for h, seq in fasta_iter(fasta_path):
        contigs[h] = seq

    total_len = sum(len(s) for s in contigs.values())
    split_len = total_len // num_process

    cur = split_len + 1
    next_ix = 0
    out = None
    with contextlib.ExitStack() as stack:
        for h,seq in contigs.items():
            if cur > split_len and next_ix < num_process:
                if out is not None:
                    out.close()
                out = open(os.path.join(output, 'contig_{}.fa'.format(next_ix)), 'wt')
                out = stack.enter_context(out)

                cur = 0
                next_ix += 1
            out.write(f'>{h}\n{seq}\n')
            cur += len(seq)

    with LoggingPool(num_process) if num_process != 0 else LoggingPool() as pool:
        try:
            for index in range(next_ix):
                pool.apply_async(
                    prodigal,
                    args=(
                        os.path.join(output, 'contig_{}.fa'.format(index)),
                        os.path.join(output, 'contig_{}.faa'.format(index)),
                    ))
            pool.close()
            pool.join()
        except:
            sys.stderr.write(
                f"Error: Running prodigal fail\n")
            sys.exit(1)

    contig_output = os.path.join(output, 'contigs.faa')
    with open(contig_output, 'w') as f:
        for index in range(next_ix):
            f.write(open(os.path.join(output, 'contig_{}.faa'.format(index)), 'r').read())
    return contig_output

def run_fraggenescan(fasta_path, num_process, output):
    try:
        contig_output = os.path.join(output, 'contigs.faa')
        with open(contig_output + '.out', 'w') as frag_out_log:
            # We need to call FragGeneScan instead of the Perl wrapper because the
            # Perl wrapper does not handle filepaths correctly if they contain spaces
            # This binary does not handle return codes correctly, though, so we
            # cannot use `check_call`:
            subprocess.call(
                [shutil.which('FragGeneScan'),
                 '-s', fasta_path,
                 '-o', contig_output,
                 '-w', str(0),
                 '-t', 'complete',
                 '-p', str(num_process),
                 ],
                stdout=frag_out_log,
            )
    except:
        sys.stderr.write(
            f"Error: Running fraggenescan failed\n")
        sys.exit(1)
    return contig_output + '.faa'

def cal_num_bins(fasta_path, binned_length, num_process, multi_mode=False, output = None, orf_finder = 'prodigal'):
    '''Estimate number of bins from a FASTA file

    Parameters
    fasta_path: path
    binned_length: int (minimal contig length)
    num_process: int (number of CPUs to use)
    multi_mode: bool, optional (if True, treat input as resulting from concatenating multiple files)
    '''
    with tempfile.TemporaryDirectory() as tdir:
        if output is not None:
            if os.path.exists(os.path.join(output, 'markers.hmmout')):
                return get_marker(os.path.join(output, 'markers.hmmout'), fasta_path, binned_length, multi_mode, orf_finder=orf_finder)
            else:
                os.makedirs(output, exist_ok=True)
                target_dir = output
        else:
            target_dir = tdir

        run_orffinder = run_prodigal if orf_finder == 'prodigal' else run_fraggenescan
        contig_output = run_orffinder(fasta_path, num_process, tdir)

        hmm_output = os.path.join(target_dir, 'markers.hmmout')
        try:
            with open(os.path.join(tdir, 'markers.hmmout.out'), 'w') as hmm_out_log:
                subprocess.check_call(
                    ['hmmsearch',
                     '--domtblout',
                     hmm_output,
                     '--cut_tc',
                     '--cpu', str(num_process),
                     os.path.split(__file__)[0] + '/marker.hmm',
                     contig_output,
                     ],
                    stdout=hmm_out_log,
                )
        except:
            if os.path.exists(hmm_output):
                os.remove(hmm_output)
            sys.stderr.write(
                f"Error: Running hmmsearch fail\n")
            sys.exit(1)

        marker = get_marker(hmm_output, fasta_path, binned_length, multi_mode, orf_finder=orf_finder)

        return marker


def write_bins(namelist, contig_labels, output, contig_seqs,
               recluster=False, origin_label=0, minfasta = 200000):
    '''
    Write binned FASTA files

    Returns: DataFrame with information on the bins
    '''
    from collections import defaultdict
    import pandas as pd

    res = defaultdict(list)
    for label, name in zip(contig_labels, namelist):
        if label != -1:
            res[label].append(name)

    os.makedirs(output, exist_ok=True)

    written = []
    for label, contigs in res.items():
        sizes = [len(contig_seqs[contig]) for contig in contigs]
        whole_bin_bp = sum(sizes)

        if whole_bin_bp >= minfasta:
            ofname = f'bin.{label}.fa' if not recluster \
                    else f'recluster_{origin_label}.bin.{label}.fa'
            ofname = os.path.join(output, ofname)
            n50, l50 = n50_l50(sizes)
            written.append([ofname, whole_bin_bp, len(contigs), n50, l50])
            with atomic_write(ofname, overwrite=True) as ofile:
                for contig in contigs:
                    ofile.write(f'>{contig}\n{contig_seqs[contig]}\n')
    return pd.DataFrame(written, columns=['filename', 'nbps', 'nr_contigs', 'N50', 'L50'])



def set_random_seed(seed):
    import torch
    import numpy as np
    torch.manual_seed(seed)
    torch.cuda.manual_seed(seed)
    torch.cuda.manual_seed_all(seed)
    random.seed(seed)
    np.random.seed(seed)

def process_fasta(fasta_path, ratio):
    """
    Returns

    binned_short: whether to include short contigs
    must_link_threshold: threshold to break up contigs
    contigs: dictionary ID -> contig sequence
    """
    whole_contig_bp = 0
    contig_bp_2500 = 0
    contig_length_list = []
    contig_dict = {}

    for h, seq in fasta_iter(fasta_path):
        if 1000 <= len(seq) <= 2500:
            contig_bp_2500 += len(seq)
        contig_length_list.append(len(seq))
        whole_contig_bp += len(seq)
        contig_dict[h] = seq

    binned_short = contig_bp_2500 / whole_contig_bp < ratio
    must_link_threshold = get_must_link_threshold(contig_length_list)
    return binned_short, must_link_threshold, contig_dict


def split_data(data, sample, separator, is_combined = True):
    """
    split data according their sample in multi-sample binning
    """
    import numpy as np
    part_data = data[data['contig_name'].str.contains(
        sample + separator, regex=False)]
    part_data = part_data.set_index('contig_name')
    part_data.index.name = None
    part_data.index = [ix.split(separator)[1] for ix in part_data.index]
    if is_combined:
        abun_scale = (part_data.mean() / 100).apply(np.ceil) * 100
        part_data = part_data.div(abun_scale)
    return part_data

def get_model_path(env):
    envn = env.lower().replace('-', '_')
    known_environments = [
            'human_gut',
            'dog_gut',
            'ocean',
            'soil',
            'cat_gut',
            'human_oral',
            'mouse_gut',
            'pig_gut',
            'built_environment',
            'wastewater',
            'chicken_caecum',
            'global',
            ]
    if envn in known_environments:
        return os.path.join(os.path.split(__file__)[0], f'{envn}_model.h5')
    else:
        sys.stderr.write(
            f"Error: Expected environment '{env}' does not exist (known ones are {', '.join(known_environments)})\n")
        sys.exit(1)

def concatenate_fasta(fasta_files, min_length, output, separator):
    cfasta = os.path.join(output, 'concatenated.fa')
    with open(cfasta, 'wt') as concat_out:
        for fasta in fasta_files:
            sample_name = os.path.basename(fasta).split('.')[0]
            for h, seq in fasta_iter(fasta):
                if separator in h:
                    sys.stderr.write(
                        f"Error: the header of the contig '{h}' contains the separator ('{separator}'), please choose another separator.\n")
                    sys.exit(1)
                if len(seq) >= min_length:
                    header = f'{sample_name}{separator}{h}'
                    concat_out.write(f'>{header}\n{seq}\n')

def n50_l50(sizes):
    '''Computes N50 & L50 for a set of contig sizes

    Returns (n50, l50)
    '''
    import numpy as np
    s = np.array(sizes)
    s.sort()
    s = s[::-1]
    l50 = np.where(np.cumsum(s)>=s.sum()//2)[0][0]
    n50 = s[l50]
    return n50, l50+1




from SemiBin.utils import get_must_link_threshold, get_marker, split_data, n50_l50
from hypothesis import given, strategies as st
from io import StringIO
import numpy as np

def slow_get_must_link_threshold(contig_len):
    """
    calculate the threshold length for must link breaking up
    """
    basepair_sum = 0
    threshold = 0
    whole_len = np.sum(contig_len)
    contig_len.sort(reverse=True)
    index = 0
    while(basepair_sum / whole_len < 0.98):
        basepair_sum += contig_len[index]
        threshold = contig_len[index]
        index += 1
    return np.clip(threshold, 4000, None)


@given(lens=st.lists(st.integers(min_value=1, max_value=(1<<32)), min_size=1))
def test_get_must_link_threshold(lens):
    assert get_must_link_threshold(lens) == slow_get_must_link_threshold(lens)


HMMOUT_TEST_SINGLE = '''#                                                                            --- full sequence --- -------------- this domain -------------   hmm coord   ali coord   env coord
# target name        accession   tlen query name           accession   qlen   E-value  score  bias   #  of  c-Evalue  i-Evalue  score  bias  from    to  from    to  from    to  acc description of target
#------------------- ---------- ----- -------------------- ---------- ----- --------- ------ ----- --- --- --------- --------- ------ ----- ----- ----- ----- ----- ----- ----- ---- ---------------------
k119_224042_1 -            122 TIGR00855            TIGR00855    125     3e-39  129.7   6.7   1   1   4.4e-42   3.3e-39  129.6   6.7     2   125     2   122     1   122 0.92 -
#
# Program:         hmmsearch
# Version:         3.3.2 (Nov 2020)
# Pipeline mode:   SEARCH
# Query file:      /home/luispedro/.miniconda3/envs/py3.9/lib/python3.9/site-packages/SemiBin-0.4.0-py3.9.egg/SemiBin/marker.hmm
# Target file:     restart/bin.50.fa.frag.faa
# Option settings: hmmsearch --domtblout restart/bin.50.fa.hmmout --cut_tc --cpu 4 /home/luispedro/.miniconda3/envs/py3.9/lib/python3.9/site-packages/SemiBin-0.4.0-py3.9.egg/SemiBin/marker.hmm restart/bin.50.fa.frag.faa
# Current dir:     /home/luispedro/Sync/work/SemiBin.rec
# [ok]
'''

def test_get_marker():
    rs = get_marker('test/data/bin.230.fa.hmmout', orf_finder='fraggenescan')
    assert rs == ['k119_487808', 'k119_606328']
    rs = get_marker('test/data/bin.96.fa.hmmout', orf_finder='fraggenescan')
    assert rs == ['k119_268294', 'k119_337646']

    # This caused a crash in a previous version of SemiBin
    rs = get_marker(StringIO(HMMOUT_TEST_SINGLE), orf_finder='prodigal')
    assert rs == ['k119_224042']

def test_get_marker_multiple():
    rs = get_marker('test/data/concatenated.hmmout.gz', multi_mode=True, orf_finder='fraggenescan')
    # This was computed by running the previous version
    assert rs == {
         'bin000005': ['k119_46822'],
         'bin000051': ['k119_368830', 'k119_419149'],
         'bin000052': ['k119_314249', 'k119_686387'],
         'bin000056': ['k119_266899', 'k119_447637'],
         'bin000057': ['k119_464254'],
         'bin000078': ['k119_162461', 'k119_278484', 'k119_678785', 'k119_684640'],
         'bin000085': ['k119_232333', 'k119_549069'],
         'bin000110': ['k119_276145', 'k119_50857']}


def test_split_data():
    '''Regression test for #68

    https://github.com/BigDataBiology/SemiBin/issues/68
    '''
    import pandas as pd
    data = pd.DataFrame([[3.02367003e-02, 3.02367003e-02, 3.02367003e-02, 3.02367003e-02,
            3.02367003e-02, 3.02367003e-02, 3.02367003e-02, 3.02367003e-02,
            3.02367003e-02, 3.02367003e-02],
           [3.02367003e-02, 3.02367003e-02, 3.02367003e-02, 3.02367003e-02,
            3.02367003e-02, 3.02367003e-02, 3.02367003e-02, 3.02367003e-02,
            3.02367003e-02, 3.02367003e-02],
           [3.02367003e-02, 3.02367003e-02, 3.02367003e-02, 3.02367003e-02,
            3.02367003e-02, 3.02367003e-02, 3.02367003e-02, 3.02367003e-02,
            3.02367003e-02, 3.02367003e-02],
           [3.02367003e-02, 3.02367003e-02, 3.02367003e-02, 3.02367003e-02,
            3.02367003e-02, 3.02367003e-02, 3.02367003e-02, 3.02367003e-02,
            3.02367003e-02, 3.02367003e-02],
           [1.00000000e-05, 1.00000000e-05, 1.00000000e-05, 1.00000000e-05,
            1.00000000e-05, 1.00000000e-05, 1.00000000e-05, 1.00000000e-05,
            1.00000000e-05, 1.00000000e-05]],
           index=[
               'S1_2341+340_METAG:g1k_5_1',
               'S1_2341+340_METAG:g2k_4_1',
               'S2_2341+340_METAG:g1k_2_1',
               'S2_2341+340_METAG:g1k_3_1',
               'S2_2341+340_METAG:g1k_5_1'],
            columns=[
               'sorted10.sam_cov',
               'sorted1.sam_cov',
               'sorted2.sam_cov',
               'sorted3.sam_cov',
               'sorted4.sam_cov',
               'sorted5.sam_cov',
               'sorted6.sam_cov',
               'sorted7.sam_cov',
               'sorted8.sam_cov',
               'sorted9.sam_cov'],
            ).reset_index().rename(columns={'index':'contig_name'})

    split1 = split_data(data, 'S1_2341+340_METAG', ':')
    assert len(split1) == 2
    split2 = split_data(data, 'S2_2341+340_METAG', ':')
    assert len(split2) == 3


@given(sizes=st.lists(st.integers(min_value=2, max_value=(1<<32)), min_size=1))
def test_n50_l50(sizes):
    n50,l50 = n50_l50(sizes)
    sizes = np.array(sizes)
    sizes.sort()
    sizes = sizes[::-1]
    assert l50 <= len(sizes)
    assert np.sum(sizes[:l50]) >= np.sum(sizes)//2
    assert np.sum(sizes[:l50-1]) < np.sum(sizes)//2
    assert n50 == sizes[l50-1]




from SemiBin.main import binning
from SemiBin.fasta import fasta_iter
import os
import pytest
import logging
import pandas as pd

def test_bin():
    logger = logging.getLogger('SemiBin')
    logger.setLevel(logging.INFO)
    sh = logging.StreamHandler()
    sh.setFormatter(logging.Formatter('%(asctime)s - %(message)s'))
    logger.addHandler(sh)

    contig_dict = {h:seq for h,seq in fasta_iter('test/bin_data/input.fasta')}

    os.makedirs('output_bin',exist_ok=True)
    binning(num_process=1,
            data='test/bin_data/data.csv',
            max_edges=20,
            max_node=1,
            minfasta=0,
            logger=logger,
            output='output_bin',
            binned_length=1000,
            device='cpu',
            contig_dict=contig_dict,
            model_path='test/bin_data/model.h5',
            recluster=True,
            random_seed=None,
            environment=None,
            )

    assert len(os.listdir('output_bin/output_prerecluster_bins')) > 0
    assert len(os.listdir('output_bin/output_recluster_bins')) > 0



import os
from os import path
import shutil
import sys
from time import sleep

def download_GTDB_to(logger, GTDB_dir):
    """
    Download GTDB.
    logger: logger
    GTDB_dir: where to store the data
    """
    import requests
    import tarfile
    import hashlib
    logger.info(f'Downloading GTDB to {GTDB_dir}.  It will take a while..')
    os.makedirs(GTDB_dir, exist_ok=True)

    download_path = os.path.join(GTDB_dir, 'GTDB_v95.tar.gz')

    # This allows one to bypass the actual downloading for testing (otherwise
    # tests take a few hours!)
    if path.exists('__SemiBin__internal_test_GTDB_v95.tar.gz'):
        logger.warning(f'Found __SemiBin__internal_test_GTDB_v95.tar.gz')
        logger.warning(f'Will copy it to {download_path} and use it as GTDB tarball')
        shutil.copy2('__SemiBin__internal_test_GTDB_v95.tar.gz', download_path)
        md5_ok = True
    else:
        download_url = 'https://zenodo.org/record/4751564/files/GTDB_v95.tar.gz?download=1'
        with requests.get(download_url, stream=True) as r:
            m = hashlib.md5()
            with open(download_path, 'wb') as f:
                while True:
                    data = r.raw.read(4096)
                    if not data:
                        break
                    f.write(data)
                    m.update(data)
        logger.info('Download finished. Checking MD5...')
        md5_ok = m.hexdigest() == '4a70301c54104e87d5615e3f2383c8b5'
    if md5_ok:
        try:
            tar = tarfile.open(download_path, "r:gz")
            file_names = tar.getnames()
            for file_name in file_names:
                tar.extract(file_name, GTDB_dir)
            tar.close()
        except Exception:
            sys.stderr.write(
                f"Error: cannot untar the file.")
            sys.exit(1)

        os.remove(download_path)
    else:
        os.remove(download_path)
        sys.stderr.write(
            f"Error: MD5 check failed. Downloading GTDB database failed (Please check the internet connections or storage space), removing '{download_path}'.\n")
        sys.exit(1)


def find_or_download_gtdb(logger, GTDB_reference_dir, force):
    '''Find or (lazily) download GTDB

    logger: logger
    GTDB_reference_dir: str
        where to store the database. Can be None, in which case, the default will be used
    force: bool
        If a database is found and `force` is true, then remove it and redownload it.
    '''
    using_default_dir = False
    if GTDB_reference_dir is None:
        cache_home = os.environ.get('XDG_CACHE_HOME')
        if cache_home is None:
            cache_home = path.join(os.environ['HOME'], '.cache')

        GTDB_reference_dir = os.path.join(
                cache_home,
                'SemiBin',
                'mmseqs2-GTDB')
        using_default_dir = True
    GTDB_reference = path.join(GTDB_reference_dir, 'GTDB')
    if path.exists(GTDB_reference):
        if force:
            logger.warning(f'Found existing directory {GTDB_reference_dir}, but `--force` was used and will be removed before re-downloading GTDB')
            if sys.stdin and sys.stdin.isatty():
                print("Waiting 5 seconds to remove...")
                for i in range(5):
                    print(f'{5-i}...')
                    sleep(1)
            os.unlink(GTDB_reference)
        else:
            logger.info(f'Found GTDB directory: `{GTDB_reference_dir}`.')
    if not path.exists(GTDB_reference):
        download_GTDB_to(logger, GTDB_reference_dir)
        if using_default_dir:
            with open(path.join(GTDB_reference_dir, 'CACHEDIR.TAG'), 'wt') as cachedir_tag:
                cachedir_tag.write('Signature: 8a477f597d28d172789f06886806bc55\n')
                cachedir_tag.write('# This file is a cache directory tag created SemiBin.\n')
                cachedir_tag.write('# For information about cache directory tags, see:\n')
                cachedir_tag.write('# http://www.brynosaurus.com/cachedir/\n')

    return GTDB_reference



def fasta_iter(fname, full_header=False):
    '''Iterate over a (possibly gzipped) FASTA file

    Parameters
    ----------
    fname : str
        Filename.
            If it ends with .gz, gzip format is assumed
            If .bz2 then bzip2 format is assumed
            if .xz, then lzma format is assumerd
    full_header : boolean (optional)
        If True, yields the full header. Otherwise (the default), only the
        first word

    Yields
    ------
    (h,seq): tuple of (str, str)
    '''
    header = None
    chunks = []
    if fname.endswith('.gz'):
        import gzip
        op = gzip.open
    elif fname.endswith('.bz2'):
        import bz2
        op = bz2.open
    elif fname.endswith('.xz'):
        import lzma
        op = lzma.open
    else:
        op = open
    with op(fname, 'rt') as f:
        for line in f:
            if line[0] == '>':
                if header is not None:
                    yield header,''.join(chunks)
                line = line[1:].strip()
                if not line:
                    header = ''
                elif full_header:
                    header = line.strip()
                else:
                    header = line.split()[0]
                chunks = []
            else:
                chunks.append(line.strip())
        if header is not None:
            yield header, ''.join(chunks)




import torch
from torch.utils.data import DataLoader
import os
from torch.optim import lr_scheduler
import sys
from .semi_supervised_model import Semi_encoding_single, Semi_encoding_multiple, feature_Dataset

def loss_function(embedding1, embedding2, label):
    relu = torch.nn.ReLU()
    d = torch.norm(embedding1 - embedding2, p=2, dim=1)
    square_pred = torch.square(d)
    margin_square = torch.square(relu(1 - d))
    supervised_loss = torch.mean(
        label * square_pred + (1 - label) * margin_square)
    return supervised_loss

def train_self(out, logger, datas, data_splits, is_combined=True,
          batchsize=2048, epoches=15, device=None, num_process = 8, mode = 'single'):
    """
    Train model from one sample(--mode single) or several samples(--mode several)
    """
    from tqdm import tqdm
    import pandas as pd
    from sklearn.preprocessing import normalize
    import numpy as np

    train_data = pd.read_csv(datas[0], index_col=0).values
    if not is_combined:
        train_data_input = train_data[:, 0:136]
    else:
        train_data_input = train_data

    torch.set_num_threads(num_process)

    logger.info('Training model...')

    if not is_combined:
        model = Semi_encoding_single(train_data_input.shape[1]).to(device)
    else:
        model = Semi_encoding_multiple(train_data_input.shape[1]).to(device)

    optimizer = torch.optim.Adam(model.parameters(), lr=1e-3)
    scheduler = lr_scheduler.StepLR(optimizer, step_size=1, gamma=0.9)

    for epoch in tqdm(range(epoches)):
        for data_index in range(len(datas)):
            if epoch == 0:
                logger.info('Generate training data of {}:'.format(data_index))

            data = pd.read_csv(datas[data_index], index_col=0)
            data.index = data.index.astype(str)
            data_split = pd.read_csv(data_splits[data_index], index_col=0)

            if mode == 'several':
                if data.shape[1] != 138 or data_split.shape[1] != 136:
                    sys.stderr.write(
                        f"Error: training mode with several only used in single-sample binning!\n")
                    sys.exit(1)

            train_data = data.values
            train_data_must_link = data_split.values

            if not is_combined:
                train_data_input = train_data[:, 0:136]
                train_data_split_input = train_data_must_link
            else:
                train_data_input = train_data
                train_data_split_input = train_data_must_link

                if train_data_input.shape[1] - 136 > 20:
                    train_data_kmer = train_data_input[:,0:136]
                    train_data_depth = train_data_input[:, 136:len(data.values[0])]
                    train_data_depth = normalize(train_data_depth, axis=1, norm='l1')
                    train_data_input = np.concatenate((train_data_kmer, train_data_depth), axis=1)

                    train_data_split_kmer = train_data_split_input[:,0:136]
                    train_data_split_depth = train_data_split_input[:, 136:len(data.values[0])]
                    train_data_split_depth = normalize(train_data_split_depth, axis=1, norm='l1')
                    train_data_split_input = np.concatenate((train_data_split_kmer, train_data_split_depth), axis = 1)


            data_length = len(train_data_input)

            n_samples = min(len(train_data_split_input) * 1000 // 2, 4000000)
            indices1 = np.random.choice(data_length, size=n_samples)
            indices2 = indices1 + 1 + np.random.choice(data_length - 1,
                                                       size=n_samples)
            indices2 %= data_length
            train_labels = [0] * n_samples

            train_input_1 = []
            train_input_2 = []

            # must link from breaking up

            for index1, index2 in zip(indices1, indices2):
                train_input_1.append(train_data_input[index1])
                train_input_2.append(train_data_input[index2])

            for i in range(0, len(train_data_must_link), 2):
                train_input_1.append(train_data_split_input[i])
                train_input_2.append(train_data_split_input[i + 1])
                train_labels.append(1)

            if epoch == 0:
                logger.debug(
                    'Number of must-link pairs:{}'.format(train_labels.count(1)))
                logger.debug(
                    'Number of cannot-link pairs:{}'.format(train_labels.count(0)))

            dataset = feature_Dataset(train_input_1, train_input_2,
                                      train_labels)
            train_loader = DataLoader(
                dataset=dataset,
                batch_size=batchsize,
                shuffle=True,
                num_workers=0)

            for train_input1, train_input2, train_label in train_loader:
                model.train()
                train_input1 = train_input1.to(device)
                train_input2 = train_input2.to(device)
                train_label = train_label.to(device)
                embedding1, embedding2 = model.forward(
                    train_input1.float(), train_input2.float())
                # decoder1, decoder2 = model.decoder(embedding1, embedding2)
                optimizer.zero_grad()
                supervised_loss = loss_function(embedding1.double(), embedding2.double(), train_label.double())
                supervised_loss = supervised_loss.to(device)
                supervised_loss.backward()
                optimizer.step()
        scheduler.step()

    logger.info('Training finished.')
    torch.save(model, os.path.join(out, 'model.h5'))

    return model



from torch.serialization import SourceChangeWarning
import warnings
warnings.filterwarnings("ignore", category=SourceChangeWarning)
from .semibin_version import __version__ as ver
import argparse
import logging
import os
import subprocess
from .atomicwrite import atomic_write
import shutil
import sys
from itertools import groupby

from .utils import validate_normalize_args, get_must_link_threshold, generate_cannot_link, \
    set_random_seed, process_fasta, split_data, get_model_path
from .gtdb import find_or_download_gtdb
from .generate_coverage import generate_cov, combine_cov
from .generate_kmer import generate_kmer_features_from_fasta
from .cluster import cluster
from .fasta import fasta_iter
from .error import LoggingPool


def parse_args(args):
    parser = argparse.ArgumentParser(formatter_class=argparse.ArgumentDefaultsHelpFormatter,
                                     description='Semi-supervised siamese neural '
                                                 'network for metagenomic binning')

    parser.version = ver

    parser.add_argument('-v',
                        '-V',
                        '--version',
                        action='version',
                        help='Print the version number')
    parser.add_argument('--verbose',
                       required=False,
                       help='Verbose output',
                       dest='verbose',
                       action='store_true', )

    subparsers = parser.add_subparsers(title='SemiBin subcommands',
                                       dest='cmd',
                                       metavar='')

    single_easy_bin = subparsers.add_parser('single_easy_bin',
                                            help='Bin contigs (single or co-assembly) using one command.')

    multi_easy_bin = subparsers.add_parser('multi_easy_bin',
                                            help='Bin contigs (multi-sample mode) using one command.')

    generate_cannot_links = subparsers.add_parser('generate_cannot_links', aliases=['predict_taxonomy',],
                                             help='Run the contig annotation using mmseqs '
                                                  'with GTDB reference genome and generate '
                                                  'cannot-link file used in the semi-supervised deep learning model training. '
                                                  'This will download the GTDB database if not downloaded before.')

    generate_sequence_features_single = subparsers.add_parser('generate_sequence_features_single', aliases=['generate_sequence_features_single'],
                                            help='Generate sequence features (kmer and abundance) as training data'
                                                  ' for semi-supervised deep learning model training (single or co-assembly mode).'
                                                  ' This will produce the data.csv and data_split.csv files.'
                                                  )


    generate_sequence_features_multi = subparsers.add_parser('generate_sequence_features_multi', aliases=['generate_sequence_features_multi'],
                                            help='Generate sequence features (kmer and abundance) as training data'
                                                  ' for semi-supervised deep learning model training (multi-sample mode).'
                                                  ' This will produce the data.csv and data_split.csv files.'
                                                  )

    download_GTDB = subparsers.add_parser('download_GTDB', help='Download GTDB reference genomes.')

    check_install = subparsers.add_parser('check_install', help = 'Check whether required dependencies are present.')

    concatenate_fasta = subparsers.add_parser('concatenate_fasta', help = 'concatenate fasta files for multi-sample binning')

    concatenate_fasta.add_argument('-i', '--input-fasta',
                   required=True,
                   nargs='*',
                   help='Path to the input fasta file.',
                   dest='contig_fasta',
                   default=None, )

    concatenate_fasta.add_argument('-o', '--output',
                        required=True,
                        help='Output directory (will be created if non-existent)',
                        dest='output',
                        default=None
                        )

    concatenate_fasta.add_argument('-m', '--min-len',
                        required=False,
                        type=int,
                        help='Discard sequences below this length (default:0)',
                        default=0,
                        dest='min_length')

    generate_sequence_features_single.add_argument('--kmer',
                                                   required=False,
                                                   help='Just output data.csv with k-mer features.',
                                                   dest='kmer',
                                                   action='store_true',)

    training = subparsers.add_parser('train',
                                    help='Train the model.')

    training_self = subparsers.add_parser('train_self',
                                          help = 'Train the model with self-supervised learning')

    binning = subparsers.add_parser('bin',
                                    help='Group the contigs into bins.')

    download_GTDB.add_argument('-f', '--force',
                            required=False,
                            help='Redownload GTDB even if files are found',
                            dest='force',
                            action='store_true',
                            default=None)

    for p in [single_easy_bin, multi_easy_bin, generate_sequence_features_single, generate_sequence_features_multi, check_install, concatenate_fasta, training, binning, training_self]:
        p.add_argument('--verbose',
                            required=False,
                            help='Verbose output',
                            dest='verbose',
                            action='store_true', )

    for p in [training, training_self]:
        p.add_argument('--data',
                             required=True,
                             nargs='*',
                             help='Path to the input data.csv file.',
                             dest='data',
                             default=None,
                             )
        p.add_argument('--data-split',
                             required=True,
                             nargs='*',
                             help='Path to the input data_split.csv file.',
                             dest='data_split',
                             default=None,
                             )

        p.add_argument('--batch-size',
                              required=False,
                              type=int,
                              help='Batch size used in the training process (Default: 2048).',
                              dest='batchsize',
                              default=2048)

        p.add_argument('--mode',
                              required=True,
                              type=str,
                              help='[single/several] Train models from one (single) or more samples (several). '
                                   'In `several` mode, you must provide data, data_split, cannot, and fasta files for corresponding samples in the same order. '
                                   'Note: You can only use `several` mode when performing single-sample binning. Training from several samples with multi-sample binning is not supported.',
                              dest='mode',
                              default='single')

        p.add_argument('-o', '--output',
                              required=True,
                              help='Output directory (will be created if non-existent)',
                              dest='output',
                              default=None,
                              )

    training.add_argument('-c', '--cannot-link',
                         required=True,
                          nargs='*',
                         help='Path to the input cannot link file. '
                         'The file format: `contig_1,contig_2` '
                         '(one row for each cannot link constraint).',
                         dest='cannot_link',
                         default=None)

    training.add_argument('--epochs', '--epoches', # epoches is kept for backwards compatibilty
                       required=False,
                       type=int,
                       help='Number of epochs used in the training process (Default: 20).',
                       dest='epoches',
                       default=20)

    training_self.add_argument('--epochs', '--epoches', # epoches is kept for backwards compatibilty
                       required=False,
                       type=int,
                       help='Number of epochs used in the training process (Default: 15).',
                       dest='epoches',
                       default=15)

    training.add_argument('-i', '--input-fasta',
                   required=True,
                   nargs='*',
                   help='Path to the input fasta file.',
                   dest='contig_fasta',
                   default=None, )

    binning.add_argument('--data',
                         required=True,
                         help='Path to the input data.csv file.',
                         dest='data',
                         default=None,)

    binning.add_argument('--max-edges',
                   required=False,
                   type=int,
                   help='The maximum number of edges that can be connected to one contig (Default: 200).',
                   dest='max_edges',
                   default=200)
    binning.add_argument('--max-node',
                   required=False,
                   type=float,
                   dest='max_node',
                   default=1,
                   help='Fraction of contigs that considered to be binned (should be between 0 and 1; default: 1).')

    binning.add_argument('--model',
                         required=False,
                         type=str,
                         dest='model_path',
                         default=None,
                         help='Path to the trained semi-supervised deep learning model.')

    for p in [single_easy_bin, multi_easy_bin, training, binning]:
        p.add_argument('--orf-finder',
                       required=False,
                       type=str,
                       help='ORF finder used to estimate the number of bins (prodigal/fraggenescan)',
                       dest='orf_finder',
                       default='prodigal')

    for p in [single_easy_bin, binning]:
        p.add_argument('--depth-metabat2',
                       required=False,
                       type=str,
                       help='depth file generated by metabat2 (only used with single-sample binning)',
                       dest='depth_metabat2',
                       default=None,
                       )

    for p in [single_easy_bin, multi_easy_bin, generate_cannot_links, training, binning, generate_sequence_features_single, generate_sequence_features_multi, training_self]:
        p.add_argument('--tmpdir',
                       required=False,
                       type=str,
                       help='option to set temporary directory',
                       dest='tmp_output',
                       default=None,
                       )

    for p in [training, generate_cannot_links, binning, single_easy_bin, multi_easy_bin, generate_sequence_features_single, generate_sequence_features_multi, training_self]:
        p.add_argument('-p', '--processes', '-t', '--threads',
                   required=False,
                   type=int,
                   help='Number of CPUs used (pass the value 0 to use all CPUs, default: 0)',
                   dest='num_process',
                   default=0,
                   metavar=''
                   )

    for p in [single_easy_bin, binning]:
        p.add_argument('--environment',
                       required=False,
                       help='Environment for the built-in model (available choices: human_gut/dog_gut/ocean/soil/cat_gut/human_oral/mouse_gut/pig_gut/built_environment/wastewater/chicken_caecum/global).',
                       dest='environment',
                       default=None,
                       )

    for p in [single_easy_bin, multi_easy_bin, generate_cannot_links, generate_sequence_features_single, generate_sequence_features_multi, binning]:
        p.add_argument('-i', '--input-fasta',
                                required=True,
                                help='Path to the input fasta file.',
                                dest='contig_fasta',
                                default=None,)
        p.add_argument('-o', '--output',
                            required=True,
                            help='Output directory (will be created if non-existent)',
                            dest='output',
                            default=None,
                            )
    for p in [single_easy_bin, multi_easy_bin, generate_cannot_links, generate_sequence_features_single, generate_sequence_features_multi, binning, training]:
        p.add_argument('-m', '--min-len',
                       required=False,
                       type=int,
                       help='Minimal length for contigs in binning. '
                            'If you use SemiBin with multi steps and you use this parameter, please use this parameter consistently with all subcommands. '
                            '(Default: SemiBin chooses 1000bp or 2500bp according the ratio of the number of base pairs of contigs between 1000-2500bp).',
                       dest='min_len',
                       default=None,
                       )
        p.add_argument('--ratio',
                       required=False,
                       type=float,
                       help='If the ratio of the number of base pairs of contigs between 1000-2500 bp smaller than this value, the minimal length will be set as 1000bp, otherwise 2500bp. '
                       'Note that setting `--min-length/-m` overrules this parameter. '
                       'If you use SemiBin with multi steps and you use this parameter, please use this parameter consistently with all subcommands. '
                       '(Default: 0.05)',
                       dest='ratio',
                       default=0.05)

    for p in [single_easy_bin, generate_sequence_features_single]:
        p.add_argument('-b', '--input-bam',
                              required=False,
                              nargs='*',
                              help='Path to the input BAM(.bam)/CRAM(.cram) file. '
                                   'If using multiple samples, you can input multiple files.'
                                    'If just need k-mer features, bam file is not needed.',
                              dest='bams',
                              default=None,
                              )

    for p in [multi_easy_bin, generate_sequence_features_multi]:
        p.add_argument('-b', '--input-bam',
                              required=True,
                              nargs='*',
                              help='Path to the input BAM(.bam)/CRAM(.cram) file. '
                                   'If using multiple samples, you can input multiple files.',
                              dest='bams',
                              default=None,
                              )

    for p in [single_easy_bin, multi_easy_bin, generate_cannot_links, download_GTDB]:
        p.add_argument('-r', '--reference-db-data-dir', '--reference-db',
                            required=False,
                            help='GTDB reference storage path. (Default: $HOME/.cache/SemiBin/mmseqs2-GTDB).'
                            'If not set --reference-db and SemiBin cannot find GTDB in $HOME/.cache/SemiBin/mmseqs2-GTDB, SemiBin will download GTDB (Note that >100GB of disk space are required).',
                            dest='GTDB_reference',
                            metavar='',
                            default=None)

    for p in [single_easy_bin, generate_cannot_links, multi_easy_bin]:
        p.add_argument('--cannot-name',
                            required=False,
                            help='Name for the cannot-link file(default: cannot).',
                            dest='cannot_name',
                            default='cannot',
                            metavar=''
                            )
        p.add_argument('--taxonomy-annotation-table',
                            required=False,
                            nargs='*',
                            help='Pre-computed mmseqs2 format taxonomy TSV file to bypass mmseqs2 GTDB annotation [advanced]. '
                                 'When running with multi-sample binning, please make sure that the order of the taxonomy TSV file and the contig file (used for the combined fasta) is same.',
                            dest='taxonomy_results_fname',
                            metavar='TAXONOMY_TSV')

    for p in [binning, single_easy_bin, multi_easy_bin]:
        p.add_argument('--minfasta-kbs',
                            required=False,
                            type=int,
                            help='minimum bin size in Kbps (Default: 200).',
                            dest='minfasta_kb',
                            default=200,
                            metavar='')

        p.add_argument('--no-recluster',
                           required=False,
                           help='Do not recluster bins.',
                           dest='no_recluster',
                           action='store_true', )
        p.add_argument('--recluster',
                           required=False,
                           help='[Deprecated] Does nothing (current default is to perform clustering)',
                           dest='recluster',
                           action='store_true', )
    for p in [single_easy_bin, multi_easy_bin]:
                          
        p.add_argument('--epochs', '--epoches', # epoches is kept for backwards compatibilty
                       required=False,
                       type=int,
                       help='Number of epochs used in the training process (Default: 15).',
                       dest='epoches',
                       default=15)

        p.add_argument('--batch-size',
                       required=False,
                       type=int,
                       help='Batch size used in the training process (Default: 2048).',
                       dest='batchsize',
                       default=2048)


    for p in [single_easy_bin, multi_easy_bin]:
        p.add_argument('--max-edges',
                          required=False,
                          type=int,
                          help='The maximum number of edges that can be connected to one contig (Default: 200).',
                          dest='max_edges',
                          default=200)

        p.add_argument('--max-node',
                          required=False,
                          type=float,
                          dest='max_node',
                          default=1,
                          help='Fraction of contigs that considered to be binned (should be between 0 and 1; default: 1).')

    for p in [multi_easy_bin, generate_sequence_features_multi, concatenate_fasta]:
        p.add_argument('-s', '--separator',
                           required=False,
                           type=str,
                           help='Used when multiple samples binning to separate sample name and contig name (Default is :).',
                           dest='separator',
                           default=':',
                           metavar='')

    for p in [training, binning, single_easy_bin, multi_easy_bin, training_self]:
        p.add_argument('--random-seed',
                       required=False,
                       type=int,
                       help='Random seed. Set it to a fixed value to reproduce results across runs. The default is that the seed is set by the system and .',
                       dest='random_seed',
                       default=None,
                       )

    for p in [generate_cannot_links, generate_sequence_features_single, generate_sequence_features_multi, single_easy_bin, multi_easy_bin]:
        p.add_argument('--ml-threshold',
                       required=False,
                       type=int,
                       help='Length threshold for generating must-link constraints. (By default, the threshold is calculated from the contig, and the default minimum value is 4,000 bp)',
                       dest='ml_threshold',
                       default=None)

    for p in [single_easy_bin, multi_easy_bin, training, binning, training_self]:
        p.add_argument('--engine',
                       required=False,
                       type=str,
                       help='device used to train the model (auto/gpu/cpu, auto means if SemiBin detects the gpu, SemiBin will use GPU)',
                       dest='engine',
                       default='auto')

    for p in [single_easy_bin, multi_easy_bin]:
        p.add_argument('--training-type',
                       required=False,
                       type=str,
                       help='training algorithm used to train the model (semi [default]/self)',
                       dest='training_type',
                       default='semi')

    if not args:
        parser.print_help(sys.stderr)
        sys.exit()
    args = parser.parse_args(args)
    if hasattr(args, 'no_recluster'):
        args.recluster = not args.no_recluster
    return args


def _checkback(msg):
    msg[1].info('Processed:{}'.format(msg[0]))

def check_install(verbose, orf_finder=None):
    '''Executes check_install subcommand, which checks for dependencies

    Parameters
    ----------
    verbose : boolean. If true, then prints out all the paths
    orf_finder : str, optional
    '''
    from shutil import which
    dependencies = ['bedtools', 'hmmsearch', 'mmseqs', 'FragGeneScan', 'prodigal']
    has_fgs = False
    missing_deps = False
    if verbose:
        print("Looking for dependencies...")
    for dep in dependencies:
        p = which(dep)
        if not p:
            if dep not in ['FragGeneScan', 'prodigal']:
                sys.stderr.write(
                    f"Error: {dep} does not seem to be installed!\n")
                missing_deps = True
            elif dep == 'prodigal':
                if not has_fgs:
                    sys.stderr.write(
                            'Error: neither prodigal nor FragGeneScan appear to be available!\n'
                            'At least one of them is necessary to run SemiBin\n')
                    missing_deps = True
                else:
                    if verbose or orf_finder != 'fraggenescan':
                        sys.stderr.write(
                            'Warning: prodigal does not appear to be available. You must use the `--orf-finder fraggenescan` option.\n')
                    missing_deps = True
        else:
            if dep == 'FragGeneScan':
                has_fgs = True
            if verbose:
                print(f'\t{dep:16}: {p}')
    if missing_deps:
        sys.exit(1)
    else:
        if verbose:
            print('Installation OK')

def predict_taxonomy(logger, contig_fasta, cannot_name,
                     taxonomy_results_fname, GTDB_reference, num_process,
                     binned_length, must_link_threshold, output):
    """
    Predict taxonomy using mmseqs and generate cannot-link file

    contig_fasta: contig file used
    GTDB_reference: GTDB path
    cannot_name: name for cannot-link constraint
    binned_short: threshold for contigs used in binning
    must_link_threshold: threshold of contigs for must-link constraints
    """
    import tempfile

    with tempfile.TemporaryDirectory() as tdir:
        filtered_fasta = os.path.join(tdir, 'SemiBinFiltered.fa')
        namelist = []
        num_must_link = 0
        with open(filtered_fasta, 'wt') as out:
            for h,seq in fasta_iter(contig_fasta):
                if len(seq) >= binned_length:
                    out.write(f'>{h}\n{seq}\n')
                    namelist.append(h)
                    if len(seq) >= must_link_threshold:
                        num_must_link += 1
        if taxonomy_results_fname is None:
            GTDB_reference = find_or_download_gtdb(logger, GTDB_reference, force=False)
            try:
                subprocess.check_call(
                    ['mmseqs',
                     'createdb',
                     filtered_fasta,
                     os.path.join(tdir, 'contig_DB')],
                    stdout=None,
                )
            except:
                sys.stderr.write(
                    f"Error: Running mmseqs createdb fail\n")
                sys.exit(1)
            if os.path.exists(os.path.join(output, 'mmseqs_contig_annotation')):
                shutil.rmtree(os.path.join(output, 'mmseqs_contig_annotation'))
            os.makedirs(os.path.join(output, 'mmseqs_contig_annotation'))
            try:
                subprocess.run(
                    ['mmseqs',
                     'taxonomy',
                     os.path.join(tdir, 'contig_DB'),
                     GTDB_reference,
                     os.path.join(output, 'mmseqs_contig_annotation/mmseqs_contig_annotation'),
                     tdir,
                     '--tax-lineage', '1',
                     '--threads', str(num_process),
                     ],
                    check=True,
                    stdout=None,
                )
            except:
                sys.stderr.write(
                    f"Error: Running mmseqs taxonomy fail\n")
                sys.exit(1)
            taxonomy_results_fname = os.path.join(output,
                                        'mmseqs_contig_annotation',
                                        'taxonomyResult.tsv')
            try:
                subprocess.check_call(
                    ['mmseqs',
                     'createtsv',
                     os.path.join(tdir, 'contig_DB'),
                     os.path.join(output, 'mmseqs_contig_annotation/mmseqs_contig_annotation'),
                     taxonomy_results_fname,
                     ],
                    stdout=None,
                )
            except:
                sys.stderr.write(
                    f"Error: Running mmseqs createtsv fail\n")
                sys.exit(1)

    os.makedirs(os.path.join(output, 'cannot'), exist_ok=True)
    generate_cannot_link(taxonomy_results_fname,
        namelist, num_must_link, os.path.join(output, 'cannot'), cannot_name)


def generate_sequence_features_single(logger, contig_fasta,
                         bams, binned_length,
                         must_link_threshold, num_process, output, output_kmer = False):
    """
    Generate data.csv and data_split.csv for training and clustering of single-sample and co-assembly binning mode.
    data.csv has the features(kmer and abundance) for original contigs.
    data_split.csv has the features(kmer and abundace) for contigs that are breaked up as must-link pair.

    """
    import pandas as pd

    if bams is None and output_kmer is False:
        sys.stderr.write(
            f"Error: You need to input bam files if you want to calculate coverage features.\n")
        sys.exit(1)

    if bams is not None and output_kmer is True:
        logger.info('We will only calculate k-mer features.')

    if not output_kmer:
        n_sample = len(bams)
        is_combined = n_sample >= 5
        bam_list = bams
        pool = LoggingPool(num_process) if num_process != 0 else LoggingPool()

        logger.info('Calculating coverage for every sample.')

        for bam_index in range(n_sample):
            pool.apply_async(
                generate_cov,
                args=(
                    bam_list[bam_index],
                    bam_index,
                    output,
                    must_link_threshold,
                    is_combined,
                    binned_length,
                    logger,
                    None
                ),
                callback=_checkback)
        pool.close()
        pool.join()

        for bam_index, bam_file in enumerate(bam_list):
            if not os.path.exists(os.path.join(output, '{}_data_cov.csv'.format(
                    os.path.split(bam_file)[-1] + '_{}'.format(bam_index)))):
                sys.stderr.write(
                    f"Error: Generating coverage file fail\n")
                sys.exit(1)
            if is_combined:
                if not os.path.exists(os.path.join(output, '{}_data_split_cov.csv'.format(
                        os.path.split(bam_file)[-1] + '_{}'.format(bam_index)))):
                    sys.stderr.write(
                        f"Error: Generating coverage file fail\n")
                    sys.exit(1)

        logger.debug('Start generating kmer features from fasta file.')
        kmer_whole = generate_kmer_features_from_fasta(
            contig_fasta, binned_length, 4)
        kmer_split = generate_kmer_features_from_fasta(
            contig_fasta, 1000, 4, split=True, split_threshold=must_link_threshold)

        data = kmer_whole
        data_split = kmer_split
        data.index = data.index.astype(str)

        if is_combined:
            data_cov, data_split_cov = combine_cov(output, bam_list, is_combined)
            data_split = pd.merge(data_split, data_split_cov, how='inner', on=None,
                                      left_index=True, right_index=True, sort=False, copy=True)
        else:
            data_cov = combine_cov(output, bam_list, is_combined)
            data_split = kmer_split

        data = pd.merge(data, data_cov, how='inner', on=None,
                                      left_index=True, right_index=True, sort=False, copy=True)


        with atomic_write(os.path.join(output, 'data.csv'), overwrite=True) as ofile:
            data.to_csv(ofile)

        with atomic_write(os.path.join(output, 'data_split.csv'), overwrite=True) as ofile:
            data_split.to_csv(ofile)

    else:
        logger.info('Only generating kmer features from fasta file.')
        kmer_whole = generate_kmer_features_from_fasta(
            contig_fasta, binned_length, 4)
        with atomic_write(os.path.join(output, 'data.csv'), overwrite=True) as ofile:
            kmer_whole.to_csv(ofile)


def generate_sequence_features_multi(logger, contig_fasta,
                        bams, num_process,
                        separator, ratio, min_length, ml_threshold, output):
    """
    Generate data.csv and data_split.csv for every sample of multi-sample binning mode.
    data.csv has the features(kmer and abundance) for original contigs.
    data_split.csv has the features(kmer and abundace) for contigs that are breaked up as must-link pair.
    """
    import pandas as pd
    n_sample = len(bams)
    is_combined = n_sample >= 5
    bam_list = bams
    if num_process != 0:
        pool = LoggingPool(num_process)
    else:
        pool = LoggingPool()

    # Gererate contig file for every sample
    sample_list = []
    contig_length_list = []

    os.makedirs(os.path.join(output, 'samples'), exist_ok=True)

    def fasta_sample_iter(fn):
        for h,seq in fasta_iter(fn):
            if separator not in h:
                raise ValueError(
                    f"Expected contigs to contain separator character ({separator}), found {h}")
            sample_name, contig_name = h.split(separator, 1)
            yield sample_name, contig_name, seq

    for sample_name, contigs in groupby(fasta_sample_iter(contig_fasta), lambda sn_cn_seq : sn_cn_seq[0]):
        with open(os.path.join( output, 'samples', '{}.fa'.format(sample_name)), 'wt') as out:
            for _, contig_name, seq in contigs:
                out.write(f'>{contig_name}\n{seq}\n')
                contig_length_list.append(len(seq))
        sample_list.append(sample_name)

    must_link_threshold = get_must_link_threshold(contig_length_list) if ml_threshold is None else ml_threshold

    logger.info('Calculating coverage for every sample.')

    binning_threshold = {}
    for sample in sample_list:
        if min_length is not None:
            binning_threshold[sample] = min_length
        else:
            binned_short ,_ ,_ = process_fasta(os.path.join(output, 'samples/{}.fa'.format(sample)), ratio)
            binning_threshold[sample] = 1000 if binned_short else 2500

    for bam_index in range(n_sample):
        pool.apply_async(generate_cov,
                         args=(
                             bam_list[bam_index],
                             bam_index,
                             os.path.join(output, 'samples'),
                             must_link_threshold,
                             is_combined,
                             binning_threshold,
                             logger,
                             separator,
                         ),
                         callback=_checkback)
    pool.close()
    pool.join()

    for bam_index, bam_file in enumerate(bam_list):
        if not os.path.exists(os.path.join(os.path.join(output, 'samples'), '{}_data_cov.csv'.format(
                os.path.split(bam_file)[-1] + '_{}'.format(bam_index)))):
            sys.stderr.write(
                f"Error: Generating coverage file fail\n")
            sys.exit(1)
        if is_combined:
            if not os.path.exists(os.path.join(os.path.join(output, 'samples'), '{}_data_split_cov.csv'.format(
                    os.path.split(bam_file)[-1] + '_{}'.format(bam_index)))):
                sys.stderr.write(
                    f"Error: Generating coverage file fail\n")
                sys.exit(1)

    # Generate cov features for every sample
    if is_combined:
        data_cov, data_split_cov = combine_cov(os.path.join(output, 'samples'), bam_list, is_combined)
        data_split_cov = data_split_cov.reset_index()
        columns_list = list(data_split_cov.columns)
        columns_list[0] = 'contig_name'
        data_split_cov.columns = columns_list
    else:
        data_cov = combine_cov(os.path.join(output, 'samples'), bam_list, is_combined)

    data_cov = data_cov.reset_index()
    columns_list = list(data_cov.columns)
    columns_list[0] = 'contig_name'
    data_cov.columns = columns_list

    for sample in sample_list:
        output_path = os.path.join(output, 'samples', sample)
        os.makedirs(output_path, exist_ok=True)

        part_data = split_data(data_cov, sample, separator, is_combined)
        part_data.to_csv(os.path.join(output_path, 'data_cov.csv'))

        if is_combined:
            part_data = split_data(data_split_cov, sample, separator, is_combined)
            part_data.to_csv(os.path.join(
                output_path, 'data_split_cov.csv'))

        sample_contig_fasta = os.path.join(
            output, 'samples/{}.fa'.format(sample))
        binned_length =  binning_threshold[sample]
        kmer_whole = generate_kmer_features_from_fasta(
            sample_contig_fasta, binned_length, 4)
        kmer_split = generate_kmer_features_from_fasta(
            sample_contig_fasta, 1000, 4, split=True, split_threshold=must_link_threshold)


        sample_cov = pd.read_csv(os.path.join(output_path, 'data_cov.csv'),index_col=0)
        kmer_whole.index = kmer_whole.index.astype(str)
        sample_cov.index = sample_cov.index.astype(str)
        data = pd.merge(kmer_whole, sample_cov, how='inner', on=None,
                        left_index=True, right_index=True, sort=False, copy=True)
        if is_combined:
            sample_cov_split = pd.read_csv(os.path.join(
                output_path, 'data_split_cov.csv'), index_col=0)
            data_split = pd.merge(kmer_split, sample_cov_split, how='inner', on=None,
                                  left_index=True, right_index=True, sort=False, copy=True)
        else:
            data_split = kmer_split

        with atomic_write(os.path.join(output_path, 'data.csv'), overwrite=True) as ofile:
            data.to_csv(ofile)

        with atomic_write(os.path.join(output_path, 'data_split.csv'), overwrite=True) as ofile:
            data_split.to_csv(ofile)

    return sample_list


def training(logger, contig_fasta, num_process,
             data, data_split, cannot_link, batchsize,
             epoches,  output, device, ratio, min_length, mode, orf_finder = None, training_mode = 'semi'):
    """
    Training the model

    model: [single/several]
    """
    from .semi_supervised_model import train
    from .self_supervised_model import train_self
    import pandas as pd
    binned_lengths = []

    if mode == 'single':
        logger.info('Start training from one sample.')
        data_ = pd.read_csv(data[0], index_col=0)
        col_name = data_.columns.tolist()[-1].split('_')[-1]
        is_combined = False if col_name == 'var' else True

    else:
        logger.info('Start training from multiple samples.')
        is_combined = False

    if training_mode == 'semi':
        contig_fasta_unzip = []
        for fasta_index,temp_fasta in enumerate(contig_fasta):
            if temp_fasta.endswith('.gz') or temp_fasta.endswith('.bz2') or temp_fasta.endswith('.xz'):
                temp_fasta_unzip = os.path.join(output, 'unzip_contig_{}.fa'.format(fasta_index))
                with open(temp_fasta_unzip, 'wt') as out:
                    for h,seq in fasta_iter(temp_fasta):
                            out.write(f'>{h}\n{seq}\n')
                contig_fasta_unzip.append(temp_fasta_unzip)
            else:
                contig_fasta_unzip.append(temp_fasta)

        if mode == 'single':
            if min_length is None:
                binned_short, _, _ = process_fasta(contig_fasta[0], ratio)
                binned_lengths.append(1000) if binned_short else binned_lengths.append(2500)
            else:
                binned_lengths.append(min_length)
        else:
            for contig_index in contig_fasta:
                if min_length is None:
                    binned_short, _, _ = process_fasta(contig_index, ratio)
                    binned_lengths.append(1000) if binned_short else binned_lengths.append(2500)
                else:
                    binned_lengths.append(min_length)

        model = train(
            output,
            contig_fasta_unzip,
            binned_lengths,
            logger,
            data,
            data_split,
            cannot_link,
            is_combined,
            batchsize,
            epoches,
            device,
            num_process,
            mode = mode,
            orf_finder=orf_finder)
    else:
        model = train_self(output,
                           logger,
                           data,
                           data_split,
                           is_combined,
                           batchsize,
                           epoches,
                           device,
                           num_process,
                           mode)


def binning(logger, num_process, data,
            max_edges, max_node, minfasta,
            binned_length, contig_dict, recluster,model_path,
            random_seed,output, device, environment, orf_finder = 'prodigal', depth_metabat2 = None):
    """
    Clustering the contigs to get the final bins.

    contig_dict: {contig_id:seq,...}
    recluster: if reclustering
    model_path: path to the trained model
    """
    import torch
    import pandas as pd
    logger.info('Start binning.')


    data = pd.read_csv(data, index_col=0)
    data.index = data.index.astype(str)

    if depth_metabat2 is None:
        col_name = data.columns.tolist()[-1].split('_')[-1]
        is_combined = False if col_name == 'var' else True
        n_sample = (data.shape[1] - 136) // 2 if not is_combined else (data.shape[1] - 136)
    else:
        is_combined = False
        n_sample = 1
        depth_metabat2 = pd.read_csv(depth_metabat2, sep='\t')
        depth_metabat2.set_index(['contigName'], inplace=True)
        depth_metabat2.drop(['contigLen', 'totalAvgDepth'], axis=1, inplace=True)
        depth_metabat2.index.name = None
        data = pd.merge(data, depth_metabat2, how='inner', on=None,
                 left_index=True, right_index=True, sort=False, copy=True)
        if data.shape[1] != 138:
            sys.stderr.write(
                f"Error: Depth file from Metabat2 only used in single-sample binning!\n")
            sys.exit(1)

    model_path = model_path if environment is None else get_model_path(environment)
    if environment is not None:
        if data.shape[1] != 138:
            sys.stderr.write(f"Error: provided pretrained model only used in single-sample binning!\n")
            sys.exit(1)

    if device == torch.device('cpu'):
        model = torch.load(model_path, map_location=torch.device('cpu'))
    else:
        model = torch.load(model_path).to(device)
    cluster(
        model,
        data,
        device,
        max_edges,
        max_node,
        is_combined,
        logger,
        n_sample,
        output,
        contig_dict,
        binned_length,
        num_process,
        minfasta,
        recluster,
        random_seed,
        orf_finder=orf_finder)


def single_easy_binning(args, logger, binned_length,
                        must_link_threshold,
                        contig_dict, recluster,random_seed, output, device, environment, orf_finder = 'prodigal', depth_metabat2 = None, training_type = 'semi'):
    """
    contain `generate_cannot_links`, `generate_sequence_features_single`, `train`, `bin` in one command for single-sample and co-assembly binning
    """
    logger.info('Generate training data.')
    if depth_metabat2 is None and args.bams is None:
        sys.stderr.write(
            f"Error: You need to input bam files if you want to calculate coverage features.\n")
        sys.exit(1)

    if args.bams is not None and depth_metabat2 is not None:
        logger.info('We will use abundance information from Metabat2.')

    if depth_metabat2 and environment is None:
        sys.stderr.write(
            f"Error: You need to use our provided model if you provide depth file from Metabat2.\n")
        sys.exit(1)

    generate_sequence_features_single(
        logger,
        args.contig_fasta,
        args.bams,
        binned_length,
        must_link_threshold,
        args.num_process,
        output,
        output_kmer= True if depth_metabat2 else False)

    data_path = os.path.join(output, 'data.csv')
    if not depth_metabat2:
        data_split_path = os.path.join(output, 'data_split.csv')

    if environment is None:
        if training_type == 'semi':
            logger.info('Running mmseqs and generate cannot-link file.')
            predict_taxonomy(
                logger,
                args.contig_fasta,
                args.cannot_name,
                None if args.taxonomy_results_fname is None else args.taxonomy_results_fname[0],
                args.GTDB_reference,
                args.num_process,
                binned_length,
                must_link_threshold,
                output)
            logger.info('Training model and clustering.')
            training(logger, [args.contig_fasta],
                     args.num_process, [data_path], [data_split_path],
                     [os.path.join(output, 'cannot', 'cannot.txt')],
                     args.batchsize, args.epoches, output, device, args.ratio, args.min_len,  mode='single', orf_finder=orf_finder, training_mode='semi')
        else:
            training(logger, None,
                     args.num_process, [data_path], [data_split_path],
                     None, args.batchsize, args.epoches, output, device, None, None,  mode='single', orf_finder=None, training_mode='self')

    binning(logger, args.num_process, data_path,
            args.max_edges, args.max_node, args.minfasta_kb * 1000,
            binned_length, contig_dict, recluster,
            os.path.join(output, 'model.h5') if environment is None else None,
            random_seed, output,  device, environment if environment is not None else None, orf_finder=orf_finder, depth_metabat2=depth_metabat2)


def multi_easy_binning(args, logger, recluster,
                       random_seed, output, device, orf_finder='prodigal', training_type = 'semi'):
    """
    contain `generate_cannot_links`, `generate_sequence_features_multi`, `train`, `bin` in one command for multi-sample binning
    """
    logger.info('Multi-sample binning.')
    logger.info('Generate training data.')

    sample_list = generate_sequence_features_multi(
        logger,
        args.contig_fasta,
        args.bams,
        args.num_process,
        args.separator,
        args.ratio,
        args.min_len,
        args.ml_threshold,
        output,)

    for sample_index, sample in enumerate(sample_list):
        logger.info(
            'Running mmseqs and generate cannot-link file of {}.'.format(sample))
        sample_fasta = os.path.join(
            output, 'samples', '{}.fa'.format(sample))
        sample_data = os.path.join(output, 'samples', sample, 'data.csv')
        sample_data_split = os.path.join(
            output, 'samples', sample, 'data_split.csv')

        binned_short, must_link_threshold, contig_dict = process_fasta(sample_fasta, args.ratio)

        if args.min_len is None:
            binned_length = 1000 if binned_short else 2500
        else:
            binned_length = args.min_len
        if args.ml_threshold is not None:
            must_link_threshold = args.ml_threshold
        logger.info('Training model and clustering for {}.'.format(sample))
        if training_type == 'semi':
            predict_taxonomy(
                logger,
                sample_fasta,
                sample,
                None if args.taxonomy_results_fname is None else args.taxonomy_results_fname[sample_index],
                args.GTDB_reference,
                args.num_process,
                binned_length,
                must_link_threshold,
                os.path.join(output, 'samples', sample))

            sample_cannot = os.path.join(
                output, 'samples', sample, 'cannot/{}.txt'.format(sample))
            training(logger, [sample_fasta], args.num_process,
                     [sample_data], [sample_data_split], [sample_cannot],
                     args.batchsize, args.epoches, os.path.join(output, 'samples', sample),
                     device, args.ratio, args.min_len, mode='single', orf_finder=orf_finder, training_mode='semi')
        else:
            training(logger, None, args.num_process,
                     [sample_data], [sample_data_split], None,
                     args.batchsize, args.epoches, os.path.join(output, 'samples', sample),
                     device, None, None, mode='single', orf_finder=None, training_mode='self')

        binning(logger, args.num_process, sample_data,
                args.max_edges, args.max_node, args.minfasta_kb * 1000,
                binned_length, contig_dict, recluster,
                os.path.join(output, 'samples', sample, 'model.h5'), random_seed ,
                os.path.join(output, 'samples', sample),  device, None, orf_finder=orf_finder)

    os.makedirs(os.path.join(output, 'bins'), exist_ok=True)
    for sample in sample_list:
        bin_dir_name = 'output_recluster_bins' if recluster else 'output_bins'
        for bf in os.listdir(os.path.join(output, 'samples', sample, bin_dir_name)):
            original_path = os.path.join(output, 'samples', sample, bin_dir_name, bf)
            new_file = '{0}_{1}'.format(sample, bf)
            new_path = os.path.join(output, 'bins', new_file)
            shutil.copyfile(original_path, new_path)


def main():
    import tempfile

    args = sys.argv[1:]
    args = parse_args(args)

    logger = logging.getLogger('SemiBin')
    loglevel = (logging.DEBUG if args.verbose else logging.INFO)
    logger.setLevel(loglevel)
    try:
        import coloredlogs
        coloredlogs.install(level=loglevel, logger=logger)
    except ImportError:
        sh = logging.StreamHandler()
        sh.setFormatter(logging.Formatter('[%(asctime)s] %(levelname)s: %(message)s'))
        logger.addHandler(sh)

    validate_normalize_args(logger, args)
    if args.cmd == 'check_install':
        check_install(True)

    if args.cmd not in ['download_GTDB', 'check_install']:
        out = args.output
        os.makedirs(out, exist_ok=True)

    if args.cmd in ['single_easy_bin', 'multi_easy_bin', 'train', 'bin', 'train_self']:
        import torch
        if args.engine == 'cpu':
            device = torch.device("cpu")
            logger.info('Running with CPU.')

        else:
            if torch.cuda.is_available():
                device = torch.device("cuda")
                logger.info('Running with GPU.')
            else:
                device = torch.device("cpu")
                logger.info('Did not detect GPU, using CPU.')

    if args.cmd in ['single_easy_bin', 'multi_easy_bin', 'generate_cannot_links', 'train', 'bin',
                    'generate_sequence_features_single', 'generate_sequence_features_multi', 'train_self']:
        tmp_output = args.tmp_output
        if tmp_output is not None:
            os.environ['TMPDIR'] = tmp_output
            os.makedirs(tmp_output, exist_ok=True)

    with tempfile.TemporaryDirectory() as tdir:
        if args.cmd in ['single_easy_bin', 'multi_easy_bin', 'generate_sequence_features_single', 'generate_sequence_features_multi']:
            bams_new = []
            if args.bams is not None:
                for bam in args.bams:
                    bam_split = bam.split('.')
                    if bam_split[-1] == 'cram':
                        output_bam = bam.split('/')[-1][:-4] + 'bam'
                        output_bam = os.path.join(tdir, output_bam)
                        with open(output_bam, 'wb') as cram_out:
                            subprocess.check_call(
                                ['samtools', 'view',
                                 '-bS',
                                 '-@', str(args.num_process),
                                 '-T', args.contig_fasta,
                                 bam],
                                stdout=cram_out)
                        bams_new.append(output_bam)
                    else:
                        bams_new.append(bam)
                args.bams = bams_new

        if args.cmd in ['generate_cannot_links', 'generate_sequence_features_single', 'bin','single_easy_bin']:
            binned_short, must_link_threshold, contig_dict = process_fasta(args.contig_fasta, args.ratio)
            if args.min_len is None:
                binned_length = 1000 if binned_short else 2500
            else:
                binned_length = args.min_len

        if args.cmd in ['generate_cannot_links', 'generate_sequence_features_single', 'generate_sequence_features_multi', 'single_easy_bin', 'multi_easy_bin']:
            if args.ml_threshold is not None:
                must_link_threshold = args.ml_threshold

        if args.cmd == 'download_GTDB':
            find_or_download_gtdb(logger, args.GTDB_reference, args.force)

        if args.cmd == 'generate_cannot_links':
            predict_taxonomy(
                logger,
                args.contig_fasta,
                args.cannot_name,
                taxonomy_results_fname=None if args.taxonomy_results_fname is None else args.taxonomy_results_fname[0],
                GTDB_reference=args.GTDB_reference,
                num_process=args.num_process,
                binned_length=binned_length,
                must_link_threshold=must_link_threshold,
                output=out)

        if args.cmd == 'generate_sequence_features_single':
            generate_sequence_features_single(
                logger,
                args.contig_fasta,
                args.bams,
                binned_length,
                must_link_threshold,
                args.num_process,
                out,
                args.kmer)

        if args.cmd == 'generate_sequence_features_multi':
            generate_sequence_features_multi(
                logger,
                args.contig_fasta,
                args.bams,
                args.num_process,
                args.separator,
                args.ratio,
                args.min_len,
                args.ml_threshold,
                out)

        if args.cmd == 'train':
            if args.random_seed is not None:
                set_random_seed(args.random_seed)
            training(logger, args.contig_fasta, args.num_process,
                     args.data, args.data_split, args.cannot_link,
                     args.batchsize, args.epoches, out, device, args.ratio, args.min_len, args.mode, orf_finder=args.orf_finder, training_mode='semi')

        if args.cmd == 'train_self':
            if args.random_seed is not None:
                set_random_seed(args.random_seed)
            training(logger, None, args.num_process,
                     args.data, args.data_split, None,
                     args.batchsize, args.epoches, out, device, None, None, args.mode, None, training_mode='self')


        if args.cmd == 'bin':
            if args.random_seed is not None:
                set_random_seed(args.random_seed)
            binning(logger, args.num_process, args.data, args.max_edges,
                    args.max_node, args.minfasta_kb * 1000, binned_length,
                    contig_dict, args.recluster, args.model_path, args.random_seed,out, device,
                    args.environment, orf_finder=args.orf_finder, depth_metabat2=args.depth_metabat2)


        if args.cmd == 'single_easy_bin':
            check_install(False, args.orf_finder)
            if args.random_seed is not None:
                set_random_seed(args.random_seed)
            if args.environment is not None:
                if args.depth_metabat2 is None:
                    if len(args.bams) != 1:
                        sys.stderr.write(
                            f"Error: provided pretrained model only used in single-sample binning!\n")
                        sys.exit(1)
            single_easy_binning(
                args,
                logger,
                binned_length,
                must_link_threshold,
                contig_dict,
                args.recluster,
                args.random_seed,
                out,
                device,
                args.environment,
                orf_finder=args.orf_finder,
                depth_metabat2=args.depth_metabat2,
                training_type=args.training_type)

        if args.cmd == 'multi_easy_bin':
            check_install(False, args.orf_finder)
            if args.random_seed is not None:
                set_random_seed(args.random_seed)
            multi_easy_binning(
                args,
                logger,
                args.recluster,
                args.random_seed,
                out,
                device,
                orf_finder=args.orf_finder,
                training_type=args.training_type)

        if args.cmd == 'concatenate_fasta':
            from .utils import concatenate_fasta
            concatenate_fasta(args.contig_fasta, args.min_length, out, args.separator)

        print('If you find SemiBin useful, please cite:\n Pan, S., Zhu, C., Zhao, XM. et al. A deep siamese neural network improves metagenome-assembled genomes in microbiome datasets across different environments. Nat Commun 13, 2326 (2022). https://doi.org/10.1038/s41467-022-29843-y.')

if __name__ == '__main__':
    main()



from SemiBin.main import generate_sequence_features_single, generate_sequence_features_multi
import os
import pytest
import logging
import pandas as pd

def test_generate_seq_feats_multi(tmpdir):
    logger = logging.getLogger('SemiBin')
    logger.setLevel(logging.INFO)
    sh = logging.StreamHandler()
    sh.setFormatter(logging.Formatter('%(asctime)s - %(message)s'))
    logger.addHandler(sh)

    os.makedirs(f'{tmpdir}/output_multi',exist_ok=True)
    generate_sequence_features_multi(bams=['test/multi_samples_data/input_multi_sorted1.bam',
                              'test/multi_samples_data/input_multi_sorted2.bam',
                              'test/multi_samples_data/input_multi_sorted3.bam',
                              'test/multi_samples_data/input_multi_sorted4.bam',
                              'test/multi_samples_data/input_multi_sorted5.bam',
                              'test/multi_samples_data/input_multi_sorted6.bam',
                              'test/multi_samples_data/input_multi_sorted7.bam',
                              'test/multi_samples_data/input_multi_sorted8.bam',
                              'test/multi_samples_data/input_multi_sorted9.bam',
                              'test/multi_samples_data/input_multi_sorted10.bam'],
                         num_process=1,
                         separator=':',
                         logger=logger,
                         output=f'{tmpdir}/output_multi',
                         contig_fasta='test/multi_samples_data/input_multi.fasta.xz',
                         ratio=0.05,
                         min_length=None,
                         ml_threshold=None,
                         )

    for i in range(10):
        data = pd.read_csv(f'{tmpdir}/output_multi/samples/S{i+1}/data.csv', index_col=0)
        data_split = pd.read_csv(f'{tmpdir}/output_multi/samples/S{i+1}/data_split.csv', index_col=0)
        assert data.shape == (20,146)
        assert data_split.shape == (40,146)



def test_generate_seq_feats_single(tmpdir):
    logger = logging.getLogger('SemiBin')
    logger.setLevel(logging.INFO)
    sh = logging.StreamHandler()
    sh.setFormatter(logging.Formatter('%(asctime)s - %(message)s'))
    logger.addHandler(sh)

    os.makedirs(f'{tmpdir}/output_single',exist_ok=True)
    generate_sequence_features_single(
                         bams=['test/single_sample_data/input.sorted.bam'],
                         num_process=1,
                         logger=logger,
                         output=f'{tmpdir}/output_single',
                         contig_fasta='test/single_sample_data/input.fasta',
                         binned_length=2500,
                         must_link_threshold=4000
                         )

    data = pd.read_csv(f'{tmpdir}/output_single/data.csv', index_col=0)
    data_split = pd.read_csv(f'{tmpdir}/output_single/data_split.csv', index_col=0)

    assert data.shape == (40,138)
    assert data_split.shape == (80,136)


def test_generate_seq_feats_coassembly(tmpdir):
    logger = logging.getLogger('SemiBin')
    logger.setLevel(logging.INFO)
    sh = logging.StreamHandler()
    sh.setFormatter(logging.Formatter('%(asctime)s - %(message)s'))
    logger.addHandler(sh)

    os.makedirs(f'{tmpdir}/output_coassembly',exist_ok=True)
    generate_sequence_features_single(bams=['test/coassembly_sample_data/input.sorted1.bam',
                               'test/coassembly_sample_data/input.sorted2.bam',
                               'test/coassembly_sample_data/input.sorted3.bam',
                               'test/coassembly_sample_data/input.sorted4.bam',
                               'test/coassembly_sample_data/input.sorted5.bam'],
                         num_process=1,
                         logger=logger,
                         output=f'{tmpdir}/output_coassembly',
                         contig_fasta='test/coassembly_sample_data/input.fasta',
                         binned_length=2500,
                         must_link_threshold=4000
                         )

    data = pd.read_csv(f'{tmpdir}/output_coassembly/data.csv', index_col=0)
    data_split = pd.read_csv(f'{tmpdir}/output_coassembly/data_split.csv', index_col=0)

    assert data.shape == (40,141)
    assert data_split.shape == (80,141)



from SemiBin.main import training
from SemiBin.fasta import fasta_iter
import os
import pytest
import logging
import pandas as pd

def test_train():
    logger = logging.getLogger('SemiBin')
    logger.setLevel(logging.INFO)
    sh = logging.StreamHandler()
    sh.setFormatter(logging.Formatter('%(asctime)s - %(message)s'))
    logger.addHandler(sh)

    contig_dict = {h:seq for h,seq in fasta_iter('test/train_data/input.fasta')}

    os.makedirs('output_train',exist_ok=True)
    training(contig_fasta = ['test/train_data/input.fasta'],
            num_process = 1,
            data = ['test/train_data/data.csv'],
            data_split = ['test/train_data/data_split.csv'],
            cannot_link = ['test/train_data/cannot.txt'],
            batchsize = 2048,
            epoches = 1,
            logger = logger,
            output = 'output_train',
            device = 'cpu',
            mode = 'single',
            ratio=0.05,
            min_length=None,
            training_mode = 'semi'
            )

    assert os.path.exists('output_train/model.h5')

    os.makedirs('output_train_self',exist_ok=True)
    training(contig_fasta = ['test/train_data/input.fasta'],
            num_process = 1,
            data = ['test/train_data/data.csv'],
            data_split = ['test/train_data/data_split.csv'],
            cannot_link = ['test/train_data/cannot.txt'],
            batchsize = 2048,
            epoches = 1,
            logger = logger,
            output = 'output_train_self',
            device = 'cpu',
            mode = 'single',
            ratio=0.05,
            min_length=None,
            training_mode = 'self'
            )

    assert os.path.exists('output_train_self/model.h5')



import os
import subprocess
from .atomicwrite import atomic_write

def calculate_coverage(depth_stream, bam_file, must_link_threshold, edge=75, is_combined=False,
                       contig_threshold=1000, sep=None, contig_threshold_dict=None):
    """
    depth_stream : an iterable like the output of bedtools genomecov
    bam_file : str filename
    must_link_threshold : int
    edge : int
    is_combined : bool

    """
    import pandas as pd
    import numpy as np
    from itertools import groupby

    contigs = []
    mean_coverage = []
    var = []

    split_contigs = []
    split_coverage = []

    for contig_name, lines in groupby(depth_stream, lambda ell: ell.split('\t', 1)[0]):
        lengths = []
        values = []
        for line in lines:
            line_split = line.strip().split('\t')
            length = int(line_split[2]) - int(line_split[1])
            value = int(float(line_split[3]))
            lengths.append(length)
            values.append(value)
        depth_value = np.zeros(sum(lengths), dtype=int)
        s = 0
        for ell,v in zip(lengths, values):
            depth_value[s:s+ell] = v
            s += ell

        if sep is None:
            cov_threshold = contig_threshold
        else:
            sample_name = contig_name.split(sep)[0]
            cov_threshold = contig_threshold_dict[sample_name]
        if len(depth_value) < cov_threshold:
            continue
        depth_value_ = depth_value[edge:-edge]
        mean_coverage.append(depth_value_.mean())
        var.append(depth_value_.var())
        contigs.append(contig_name)

        if is_combined:
            if len(depth_value) >= must_link_threshold:
                middle = len(depth_value_) // 2
                split_contigs.append(contig_name + '_1')
                split_contigs.append(contig_name + '_2')

                split_coverage.append(depth_value_[:middle].mean())
                split_coverage.append(depth_value_[middle:].mean())

    if is_combined:
        contig_cov = pd.DataFrame(
                { '{0}_cov'.format(bam_file): mean_coverage,
                }, index=contigs)
        split_contig_cov = pd.DataFrame(
                { '{0}_cov'.format(bam_file): split_coverage,
                }, index=split_contigs)
        return contig_cov, split_contig_cov
    else:
        return pd.DataFrame({
            '{0}_mean'.format(bam_file): mean_coverage,
            '{0}_var'.format(bam_file): var,
            }, index=contigs)


def generate_cov(bam_file, bam_index, out, threshold,
                 is_combined, contig_threshold, logger, sep = None):
    """
    Call bedtools and generate coverage file

    bam_file: bam files used
    out: output
    threshold: threshold of contigs that will be binned
    is_combined: if using abundance feature in deep learning. True: use
    contig_threshold: threshold of contigs for must-link constraints
    sep: separator for multi-sample binning
    """
    import numpy as np
    logger.debug('Processing `{}`'.format(bam_file))
    bam_name = os.path.split(bam_file)[-1] + '_{}'.format(bam_index)

    bed_p = subprocess.Popen(
        ['bedtools', 'genomecov',
         '-bga',
         '-ibam', bam_file],
        universal_newlines=True,
        stdout=subprocess.PIPE)

    if is_combined:
        contig_cov, must_link_contig_cov = calculate_coverage(bed_p.stdout, bam_file, threshold, is_combined = is_combined, sep = sep, contig_threshold = contig_threshold if sep is None else 1000, contig_threshold_dict =  contig_threshold if sep is not None else None)
        if bed_p.wait() != 0:
            logger.critical("Running `bedtools genomecov` failed. Please check your input files: SemiBin expects that they are sorted BAM files.")
            raise OSError("Failure in running bedtools")

        contig_cov = contig_cov.apply(lambda x: x + 1e-5)
        must_link_contig_cov = must_link_contig_cov.apply(lambda x: x + 1e-5)
        if sep is None:
            abun_scale = (contig_cov.mean() / 100).apply(np.ceil) * 100
            abun_split_scale = (must_link_contig_cov.mean() / 100).apply(np.ceil) * 100
            contig_cov = contig_cov.div(abun_scale)
            must_link_contig_cov = must_link_contig_cov.div(abun_split_scale)

        with atomic_write(os.path.join(out, '{}_data_cov.csv'.format(bam_name)), overwrite=True) as ofile:
            contig_cov.to_csv(ofile)

        with atomic_write(os.path.join(out, '{}_data_split_cov.csv'.format(bam_name)), overwrite=True) as ofile:
            must_link_contig_cov.to_csv(ofile)
    else:
        contig_cov = calculate_coverage(bed_p.stdout, bam_file, threshold, is_combined=is_combined, sep = sep,
                                        contig_threshold = contig_threshold if sep is None else 1000,
                                        contig_threshold_dict =  contig_threshold if sep is not None else None)
        if bed_p.wait() != 0:
            raise OSError("Failure in running bedtools")

        contig_cov = contig_cov.apply(lambda x: x + 1e-5)

        with atomic_write(os.path.join(out, '{}_data_cov.csv'.format(bam_name)), overwrite=True) as ofile:
            contig_cov.to_csv(ofile)

    return (bam_file, logger)

def combine_cov(cov_dir, bam_list, is_combined):
    """
    generate cov/cov_split for every sample in one file
    """
    import pandas as pd
    data_cov = pd.read_csv(os.path.join(cov_dir, '{}_data_cov.csv'.format(
        os.path.split(bam_list[0])[-1] + '_{}'.format(0))), index_col=0)
    if is_combined:
        data_split_cov = pd.read_csv(os.path.join(cov_dir, '{}_data_split_cov.csv'
                                    .format(os.path.split(bam_list[0])[-1] + '_{}'.format(0))), index_col=0)

    for bam_index, bam_file in enumerate(bam_list):
        if bam_index == 0:
            continue
        cov = pd.read_csv(os.path.join(cov_dir, '{}_data_cov.csv'.format(
                os.path.split(bam_file)[-1] + '_{}'.format(bam_index))),index_col=0)
        cov.index = cov.index.astype(str)
        data_cov = pd.merge(data_cov, cov, how='inner', on=None,
                            left_index=True, right_index=True, sort=False, copy=True)

        if is_combined:
            cov_split = pd.read_csv(os.path.join(cov_dir, '{}_data_split_cov.csv'.format(
                os.path.split(bam_file)[-1] + '_{}'.format(bam_index))), index_col=0)

            data_split_cov = pd.merge(data_split_cov, cov_split, how='inner', on=None,
                                      left_index=True, right_index=True, sort=False, copy=True)
    data_cov.index = data_cov.index.astype(str)
    if is_combined:
        return data_cov, data_split_cov
    else:
        return data_cov




###https://stackoverflow.com/questions/6728236/exception-thrown-in-multiprocessing-pool-not-detected

import multiprocessing
import traceback
from multiprocessing.pool import Pool


### Return error message when using multiprocessing
def error(msg, *args):
    return multiprocessing.get_logger().error(msg, *args)

class LogExceptions(object):
    def __init__(self, callable):
        self.__callable = callable

    def __call__(self, *args, **kwargs):
        try:
            result = self.__callable(*args, **kwargs)
        except Exception as e:
            error(traceback.format_exc())
            raise

        return result

class LoggingPool(Pool):
    def apply_async(self, func, args=(), kwds={}, callback=None):
        return Pool.apply_async(self, LogExceptions(func), args, kwds, callback)


from setuptools import setup

exec(compile(open('SemiBin/semibin_version.py').read(),
             'SemiBin/semibin_version.py', 'exec'))

long_description = open('README.md', encoding='utf-8').read()

setup(name='SemiBin',
      version=__version__,
      description='Metagenomic binning with semi-supervised siamese neural network',
      long_description = long_description,
      long_description_content_type = 'text/markdown',
      classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering :: Bio-Informatics',
        'Topic :: Scientific/Engineering :: Artificial Intelligence',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
      ],
      url='https://github.com/BigDataBiology/SemiBin',
      author='Shaojun Pan',
      author_email='shaojun1997777@gmail.com',
      license='MIT',
      packages = ['SemiBin'],
      install_requires=[
          'numpy',
          'tqdm',
          'pyyaml',
          'torch',
          'python-igraph',
          'pandas',
          'scikit-learn',
          'requests',
          'numexpr',
      ],
      package_data={
          'SemiBin': ['*.hmm','*.pl','*.h5']},
      zip_safe=False,
      entry_points={
            'console_scripts': ['SemiBin=SemiBin.main:main'],
      }
)




# Adapted from https://github.com/BinPro/CONCOCT/blob/develop/scripts/fasta_to_features.py
from itertools import product
from collections import OrderedDict

from .fasta import fasta_iter

def generate_feature_mapping(kmer_len):
    BASE_COMPLEMENT = {"A": "T", "T": "A", "G": "C", "C": "G"}
    kmer_hash = {}
    counter = 0
    for kmer in product("ATGC", repeat=kmer_len):
        kmer = ''.join(kmer)
        if kmer not in kmer_hash:
            kmer_hash[kmer] = counter
            rev_compl = tuple([BASE_COMPLEMENT[x] for x in reversed(kmer)])
            kmer_hash[''.join(rev_compl)] = counter
            counter += 1
    return kmer_hash, counter


def generate_kmer_features_from_fasta(
        fasta_file, length_threshold, kmer_len, split=False, split_threshold=0):
    import numpy as np
    import pandas as pd
    def seq_list():
        for h, seq in fasta_iter(fasta_file):
            if not split:
                yield h, seq
            elif len(seq) >= split_threshold:
                half = len(seq) // 2
                yield (h + '_1', seq[:half])
                yield (h + '_2', seq[half:])

    kmer_dict, nr_features = generate_feature_mapping(kmer_len)
    composition = OrderedDict()
    for h, seq in seq_list():
        if len(seq) < length_threshold:
            continue
        norm_seq = str(seq).upper()
        kmers = [kmer_dict[norm_seq[i:i+kmer_len]]
                for i in range(len(norm_seq) - kmer_len + 1)
                if norm_seq[i:i+kmer_len] in kmer_dict] # ignore kmers with non-canonical bases
        composition[h] = np.bincount(np.array(kmers, dtype=np.int64), minlength=nr_features)
    df = pd.DataFrame.from_dict(composition, orient='index', dtype=float)

    df = df.apply(lambda x: x + 1e-5)
    df = df.div(df.sum(axis=1), axis=0)
    return df



import os
import math
import shutil
import tempfile

from .utils import write_bins, cal_num_bins
from .fasta import fasta_iter

# This is the default in the igraph package
NR_INFOMAP_TRIALS = 10

def run_infomap1(g, edge_weights, vertex_weights, trials):
    return g.community_infomap(edge_weights=edge_weights, vertex_weights=vertex_weights, trials=trials)

def run_infomap(g, edge_weights, vertex_weights, num_process):
    '''Run infomap, using multiple processors (if available)'''
    if num_process == 1:
        return g.community_infomap(edge_weights=edge_weights, vertex_weights=vertex_weights, trials=NR_INFOMAP_TRIALS)
    import multiprocessing
    with multiprocessing.Pool(num_process) as p:
        rs = [p.apply_async(run_infomap1, (g, edge_weights, vertex_weights, 1))
                for _ in range(NR_INFOMAP_TRIALS)]
        p.close()
        p.join()
    rs = [r.get() for r in rs]
    best = rs[0]
    for r in rs[1:]:
        if r.codelength < best.codelength:
            best = r
    return best



def cal_kl(m, v, use_ne='auto'):
    # A naive implementation creates a lot of copies of what can
    # become large matrices
    import numpy as np

    m = np.clip(m, 1e-6, None)
    v = np.clip(v, 1.0, None)

    m1 = m.reshape(1, len(m))
    m2 = m.reshape(len(m), 1)

    v1 = v.reshape(1, len(v))
    v2 = v.reshape(len(v), 1)


    if use_ne != 'no':
        try:
            import numexpr as ne

            res = ne.evaluate(
                    '(log(v1) - log(v2))/2 + ( (m1 - m2)**2 + v2 ) / ( 2 * v1 ) - half',
                    {
                        'v1': v1,
                        'v2': v2,
                        'm1': m1,
                        'm2': m2,
                        # numexpr rules are that mixing with floats causes
                        # conversion to float64
                        # Note that this does not happen for integers
                        'half': np.float32(0.5),
                    })
            np.clip(res, 1e-6, 1-1e-6, out=res)
            return res
        except ImportError:
            if use_ne != 'auto':
                raise
    v_div = np.log(v1) - np.log(v2)
    v_div /= 2.0

    m_dif = m1 - m2
    m_dif **=2
    m_dif += v2
    m_dif /= 2 * v1

    v_div += m_dif
    v_div -= 0.5
    np.clip(v_div, 1e-6, 1-1e-6, out=v_div)
    return v_div



def cluster(model, data, device, max_edges, max_node, is_combined,
            logger, n_sample, out, contig_dict, binned_length, num_process, minfasta, recluster, random_seed, orf_finder = 'prodigal'):
    """
    Cluster contigs into bins
    max_edges: max edges of one contig considered in binning
    max_node: max percentage of contigs considered in binning
    """
    from igraph import Graph
    from sklearn.neighbors import kneighbors_graph
    from sklearn.cluster import KMeans
    from scipy import sparse
    import torch
    import pandas as pd
    import numpy as np
    train_data = data.values
    train_data_input = train_data[:, 0:136] if not is_combined else train_data

    depth = data.values[:, 136:len(data.values[0])].astype(np.float32)
    namelist = data.index.tolist()
    name2ix = {n:i for i, n in enumerate(namelist)}
    num_contigs = train_data_input.shape[0]
    with torch.no_grad():
        model.eval()
        x = torch.from_numpy(train_data_input).to(device)
        embedding = model.embedding(x.float()).detach().cpu().numpy()
        embedding_matrix = kneighbors_graph(
            embedding,
            n_neighbors=min(max_edges, train_data.shape[0] - 1),
            mode='distance',
            p=2,
            n_jobs=num_process)
        kmer_matrix = kneighbors_graph(
            train_data_input,
            n_neighbors=min(max_edges, train_data.shape[0] - 1),
            mode='distance',
            p=2,
            n_jobs=num_process)

        # We want to intersect the matrices, so we make kmer_matrix into a
        # matrix of 1.s and multiply
        kmer_matrix.eliminate_zeros()
        kmer_matrix.data.fill(1.)
        embedding_matrix = embedding_matrix.multiply(kmer_matrix)
        embedding_matrix.eliminate_zeros()
        del kmer_matrix

    np.clip(embedding_matrix.data, None, 1., out=embedding_matrix.data)
    embedding_matrix.data = 1 - embedding_matrix.data

    threshold = 0
    max_axis1 = embedding_matrix.max(axis=1).toarray()
    while threshold < 1:
        threshold += 0.05
        n_above = np.sum(max_axis1 > threshold)
        if round(n_above / num_contigs, 2) < max_node:
            break
    threshold -= 0.05

    embedding_matrix.data[embedding_matrix.data<= threshold] = 0
    embedding_matrix.eliminate_zeros()
    if not is_combined:
        logger.debug('Calculating depth matrix.')
        kl_matrix = None
        for k in range(n_sample):
            kl = cal_kl(depth[:,2*k], depth[:, 2*k + 1])
            if kl_matrix is None:
                kl *= -1
                kl_matrix = kl
            else:
                kl_matrix -= kl
        kl_matrix += n_sample
        kl_matrix /= n_sample
        embedding_matrix = embedding_matrix.multiply(kl_matrix)

    embedding_matrix.data[embedding_matrix.data <= 1e-6] = 0
    X, Y, V = sparse.find(embedding_matrix)
    # Find above diagonal positions:
    above_diag = Y > X
    X = X[above_diag]
    Y = Y[above_diag]

    edges = [(x,y) for x,y in zip(X, Y)]
    edge_weights = V[above_diag]

    logger.debug(f'Number of edges in clustering graph: {len(edges)}')

    g = Graph()
    g.add_vertices(np.arange(num_contigs))
    g.add_edges(edges)
    length_weight = np.array([len(contig_dict[name]) for name in namelist])
    result = run_infomap(g,
                edge_weights=edge_weights,
                vertex_weights=length_weight,
                num_process=num_process)
    contig_labels = np.zeros(shape=num_contigs, dtype=int)

    for i, r in enumerate(result):
        contig_labels[r] = i

    output_bin_path = os.path.join(out, 'output_prerecluster_bins' if recluster else 'output_bins')
    if os.path.exists(output_bin_path):
        shutil.rmtree(output_bin_path)
    os.makedirs(output_bin_path, exist_ok=True)

    bin_files = write_bins(namelist, contig_labels, output_bin_path, contig_dict,minfasta=minfasta)
    if not len(bin_files):
        logger.warning('No bins were created. Please check your input data.')
        return

    if not recluster:
        logger.info(f'Number of bins: {len(bin_files)}')
        bin_files.to_csv(os.path.join(out, 'bins_info.tsv'), index=False, sep='\t')

    else:
        logger.info(f'Number of bins prior to reclustering: {len(bin_files)}')
        if not is_combined:
            mean_index = [2 * temp for temp in range(n_sample)]
            depth_mean = depth[:, mean_index]
            abun_scale = np.ceil(depth_mean.mean(axis = 0) / 100) * 100
            depth_mean = depth_mean / abun_scale
            scaling = np.mean(np.abs(embedding)) / np.mean(depth_mean)
            base = 10
            weight = 2 * base * math.ceil(scaling / base)
            embedding_new = np.concatenate(
                (embedding, depth_mean * weight), axis=1)
        else:
            embedding_new = embedding

        logger.debug('Reclustering.')

        output_recluster_bin_path = os.path.join(out, 'output_recluster_bins')
        if os.path.exists(output_recluster_bin_path):
            shutil.rmtree(output_recluster_bin_path)

        os.makedirs(output_recluster_bin_path, exist_ok=True)

        with tempfile.TemporaryDirectory() as tdir:
            cfasta = os.path.join(tdir, 'concatenated.fna')
            with open(cfasta, 'wt') as concat_out:
                for ix,f in enumerate(bin_files['filename'].values):
                    for h,seq in fasta_iter(f):
                        h = f'bin{ix:06}.{h}'
                        concat_out.write(f'>{h}\n{seq}\n')
            seeds = cal_num_bins(
                cfasta,
                binned_length,
                num_process,
                multi_mode=True,
                orf_finder=orf_finder)

        outputs = []
        # The code below (iat call) relies on this
        assert bin_files.columns[0] == 'filename'
        for ix,bin_path in enumerate(bin_files['filename'].values):
            # if there are no hits, the output will be naturally empty
            seed = seeds.get(f'bin{ix:06}', [])
            num_bin = len(seed)

            if num_bin > 1:
                contig_list = [h for h,_ in fasta_iter(bin_path)]
                contig_index = [name2ix[c] for c in contig_list]
                re_bin_features = embedding_new[contig_index]

                seed_index = [name2ix[s] for s in seed]
                length_weight = np.array(
                    [len(contig_dict[name]) for name in contig_list])
                seeds_embedding = embedding_new[seed_index]
                kmeans = KMeans(
                    n_clusters=num_bin,
                    init=seeds_embedding,
                    n_init=1,
                    random_state=random_seed)
                kmeans.fit(re_bin_features, sample_weight=length_weight)
                labels = kmeans.labels_
                part = write_bins(contig_list, labels, os.path.join(out, 'output_recluster_bins'), contig_dict,
                           recluster=True, origin_label=int(bin_path.split('.')[-2]),minfasta = minfasta)
                outputs.append(part)
            else:
                ofname = shutil.copy(bin_path, os.path.join(out, 'output_recluster_bins'))
                bin_files.iat[ix, 0] = ofname
                outputs.append(bin_files[ix:ix+1])

        outputs = pd.concat(outputs)
        logger.info(f'Number of bins after reclustering: {len(outputs)}')
        outputs.to_csv(os.path.join(out, 'recluster_bins_info.tsv'), index=False, sep='\t')
    logger.info('Binning finish.')




from script.generate_cannot_link import generate_CAT
import pandas as pd

def test_CAT_script():
    cat_result = pd.read_csv('test/CAT_data/CAT.out', sep='\t')
    cat_result = cat_result[['# contig', 'classification', 'genus', 'species']]
    cat_result = cat_result.values
    _, cannot_link_genus, cannot_link_species = generate_CAT(cat_result)
    assert len(cannot_link_species) == 1
    assert len(cannot_link_genus) == 257

if __name__ == '__main__':
    test_CAT_script()




from SemiBin.generate_kmer import generate_feature_mapping, generate_kmer_features_from_fasta
from os import path
import pytest
import pandas as pd
from pandas.testing import assert_frame_equal
import numpy as np

def test_generate_feature_mapping():
    fm,n = generate_feature_mapping(4)
    assert n == len(set(fm.values()))


def test_kmer(tmpdir):
    fasta_file = path.join(tmpdir, 'contig.fna')

    with open(fasta_file, 'wt') as out_fasta:
        out_fasta.write(f'>example\nGCATGAGACAGTGTTTGCTGCAATACTCCCTGAGGCGCTTGCATATTATGAGAAGAATGCTCCCAAGGGAGAGTGCGTGATCGTTATAGAGGGAAAGAGCAGGCTTGAGATTCGGGAAGAAGAGAAAGCACAGTGGGAACAGCTTACCATTGAAGAACATATGGAACGCTATCTCTCCGGCGGTATGGAGAAAAAGGAAGCTATGAAGCGGGTGGCAAAGGATCGGGGCGTAAGTAAAAGAGATATTTATCAAGCGCTTCTTTAAACCTTCATATGAATGAAATAAAAACAGGTATCGGGTGTGTTC\n')

    kmer = generate_kmer_features_from_fasta(
        fasta_file, 0, kmer_len=1, split=False, split_threshold=0)
    assert_frame_equal(kmer, pd.DataFrame(
        np.array([0.547231, 0.452769]).reshape(1, 2), index=['example']))

    kmer_split = generate_kmer_features_from_fasta(
        fasta_file, 0, kmer_len=1, split=True, split_threshold=0)
    assert_frame_equal(kmer_split, pd.DataFrame(
        [[0.516340, 0.483660], [0.577922, 0.422078]], index=['example_1', 'example_2']))

    kmer = generate_kmer_features_from_fasta(
        fasta_file, 0, kmer_len=2, split=False, split_threshold=0)
    assert_frame_equal(kmer,
                       pd.DataFrame(np.array([0.169935,
                                              0.078431,
                                              0.147059,
                                              0.07516340680934419,
                                              0.058823542868123246,
                                              0.117647,
                                              0.143791,
                                              0.104575,
                                              0.06535948844461162,
                                              0.03921570613865813]).reshape(1, 10),
                                    index=['example']))

    kmer_split = generate_kmer_features_from_fasta(
        fasta_file, 0, kmer_len=2, split=True, split_threshold=0)
    assert_frame_equal(kmer_split,
                       pd.DataFrame([[0.131579, 0.06578949619112093, 0.177632, 0.07236843923128998, 0.03947372403044471, 0.151316, 0.157895, 0.098684, 0.07894738227145903, 0.02631583795010661],
                                    [0.202614, 0.0915032735272722, 0.117647, 0.07843138664615251, 0.07843138664615251, 0.084967, 0.130719, 0.111111, 0.05228761288391314, 0.05228761288391314]],
                                                index=['example_1', 'example_2']))

def test_kmer_with_ns(tmpdir):
    fasta_file = path.join(tmpdir, 'contig.fna')

    with open(fasta_file, 'wt') as out_fasta:
        out_fasta.write(f'>example\n')
        out_fasta.write('AAAAAAAAAAAAAAANAAAAAAAAAAAA\n')

    kmer = generate_kmer_features_from_fasta(
        fasta_file, 0, kmer_len=4, split=False, split_threshold=0)
    kmer = kmer.squeeze()
    assert len(kmer) == 136
    assert kmer.max() > 0.95

